import matplotlib.pyplot as plt
import numpy as np
from warnings import warn

from ..core.data import DNPData

dark_green = "#46812B"
light_green = "#67AE3E"
dark_grey = "#4D4D4F"
light_grey = "#A7A9AC"
orange = "#F37021"

plt.rcParams["lines.linewidth"] = 1.5
plt.rcParams["axes.prop_cycle"] = plt.cycler(
    color=[orange, dark_green, light_green, dark_grey, light_grey]
)

show = plt.show


def plot(data, *args, **kwargs):
    """Plot function for dnpdata object

    Args:
        data (dnpdata): dnpdata object for matplotlib plot function
        args: args for matplotlib plot function
        kwargs: kwargs for matplotlib plot function

    Example::

       # Plotting a dnpdata object
       dnp.dnpResults.plt.figure()
       dnp.dnpResults.plot(data)
       dnp.dnpResults.plt.show()

       # Plotting a workspace (dnpdata_collection)
       dnp.dnpResults.plt.figure()
       dnp.dnpResults.plot(ws['proc'])
       dnp.dnpResults.plt.show()

       # Plotting two curves on the same figure
       dnp.dnpResults.plt.figure()
       dnp.dnpResults.plot(ws['proc1'])
       dnp.dnpResults.plot(ws['proc2'])
       dnp.dnpResults.plt.show()

       # Plotting with some custom parameters
       dnp.dnpResults.plt.figure()
       dnp.dnpResults.plot(ws['proc'], 'k-', linewidth = 3.0, alpha = 0.5)
       dnp.dnpResults.plt.show()

    """

    if "dim" in kwargs:
        dim = kwargs.pop("dim")
    else:
        dim = data.dims[0]

    coord = data.coords[dim]

    data.unfold(dim)

    plt.plot(coord, data.values.real, *args, **kwargs)
    data.fold()


def fancy_plot(data, xlim=[], title="", showPar=False, *args, **kwargs):
    """Streamline Plot function for dnpdata objects

    This function creates streamlined plots for NMR and EPR spectra. The type of the spectrum is detected from the attribute "experiment_type" of the dnpdata object. Currently the following types are implemented: nmr_spectrum, epr_spectrum, enhancements_P, and inversion_recovery.

    Args:
        data (dnpdata): dnpdata object with values to plot
        xlim: list of limit values for plotting function
        title: string containing plot title
        showPar: boolean, toggle whether to show experiment parameters

    Returns:
        Returns formated matplotlib plot.

    Example::

        # Simply just plotting the dnpdata object
        dnp.fancy_plot(data)

        # Plot EPR spectrum from 344 mT to 354 mT, show experimental parameters
        dnp.fancy_plot(data, xlim=[344, 354], title="EPR Spectrum", showPar=True)

    """

    if "experiment_type" not in data.attrs:
        warn("experiment_type not defined in data.attrs, falling back to plot function")
        plot(data, *args, **kwargs)
        return
    elif data.attrs["experiment_type"] == None:
        warn("experiment_type is None, falling back to plot function")
        plot(data, *args, **kwargs)
        return

    if "dim" in kwargs:
        dim = kwargs.pop("dim")
    else:
        dim = data.dims[0]

    plt.grid(True)
    plt.title(title)

    if data.attrs["experiment_type"] == "nmr_spectrum":
        coord = data.coords[dim]
        data.unfold(dim)

        plt.plot(coord, data.values.real, *args, **kwargs)
        plt.xlabel("Chemical Shift $\delta$ (ppm)")
        plt.ylabel("NMR Signal Intensity (a.u.)")

        plt.xlim(max(coord), min(coord))

        if xlim != []:

            plt.xlim(xlim[1], xlim[0])

        if showPar == True:

            parameterString = "Freq: " + str(round(data.attrs["nmr_frequency"], 4))

            box_style = dict(boxstyle="round", facecolor="white", alpha=0.25)
            xmin, xmax, ymin, ymax = plt.axis()

            plt.text(xmin * 0.95, ymax / 10, parameterString, bbox=box_style)

    elif data.attrs["experiment_type"] == "epr_spectrum":
        coord = data.coords[dim]
        data.unfold(dim)

        plt.plot(coord, data.values.real, *args, **kwargs)
        plt.xlabel("Magnetic Field $B_{0}$ (mT)")
        plt.ylabel("EPR Signal Intensity (a.u.)")

        if xlim != []:
            plt.xlim(xlim[0], xlim[1])

        if showPar == True:
            SW = coord[-1] - coord[0]

            parameterString = (
                "MF: "
                + str(round(data.attrs["frequency"], 4))
                + "\nMP: "
                + str(round(data.attrs["power"], 3))
                + "\nCF: "
                + str(round(data.attrs["center_field"] / 10, 2))
                + "\nSW: "
                + str(round(SW, 2))
                + "\nMA: "
                + str(round(data.attrs["modulation_amplitude"], 2))
                + "\nNS: "
                + str(data.attrs["nscans"])
                + "\nTM: "
                + str(round(data.attrs["temperature"], 1))
            )

            box_style = dict(boxstyle="round", facecolor="white", alpha=0.25)
            xmin, xmax, ymin, ymax = plt.axis()

            plt.text(xmin * 1.001, ymin * 0.90, parameterString, bbox=box_style)

    elif data.attrs["experiment_type"] == "enhancements_P":
        coord = data.coords[dim]
        data.unfold(dim)

        plt.plot(coord, data.values.real, marker="o", fillstyle="none", *args, **kwargs)
        plt.xlabel("Microwave Power (dBm)")
        plt.ylabel("DNP Enhancement")

        if xlim != []:
            plt.xlim(xlim[0], xlim[1])

        # if showPar == True:

    elif data.attrs["experiment_type"] == "inversion_recovery":
        plt.plot(
            data.coords["t1"],
            data.values,
            marker="o",
            fillstyle="none",
            *args,
            **kwargs
        )

        plt.xlabel("Evolution Time T1 (s)")
        plt.ylabel("NMR Amplitude [a.u.]")

        if xlim != []:
            plt.xlim(xlim[0], xlim[1])

        # if showPar == True:

    else:

        plot(data)

    data.fold()



from ..core.data import DNPData
import numpy as np
from struct import unpack
import warnings
import os
import glob


def import_prospa(path, parameters_filename=None, experiment=None, verbose=False):
    """Import Kea data

    Args:
        path (str): Path to data
        parameters_filename (str):
        experiment (str): Prospa experiment, used when calculating coords from parameters
        verbose (bool): If true, prints additional information for troubleshooting

    Returns:
        dnpdata object with Kea data
    """

    if parameters_filename == None:
        parameters_filename = "acqu.par"
    extension = ""

    if os.path.isfile(path):
        path, filename = os.path.split(path)
        filename, extension = os.path.splitext(filename)
    elif os.path.isdir(path):
        filesList = glob.glob(os.path.join(path, "*.[1-4]d"))
        if len(filesList) == 0:
            raise ValueError("No binary data file in directory:", path)
        elif len(filesList) > 1:
            raise ValueError("More than one binary data file in directory:", filesList)
        else:
            data_filename = filesList[0]
            filename, extension = os.path.splitext(os.path.split(data_filename)[-1])
    else:
        raise OSError("%s not found" % path)

    try:
        attrs = import_par(os.path.join(path, parameters_filename))
    except:
        warnings.warn("No parameters file in directory")
        attrs = {}

    if extension == ".csv":
        # Import csv data
        x, data = import_csv(os.path.join(path, filename + extension))
    else:
        # Import Binary data
        x, data = import_nd(os.path.join(path, filename + extension))

    if "b1Freq" in attrs:
        nmr_frequency = attrs["b1Freq"]
        try:
            nmr_frequency = float(nmr_frequency)
        except:
            nmr_frequency = float(nmr_frequency.replace("d", ""))

        attrs["nmr_frequency"] = nmr_frequency * 1e6

    # Assume direct dimension is 1st dimension
    data_shape = np.shape(np.squeeze(data))

    dims, coords = prospa_coords(attrs, data_shape, experiment=experiment)

    kea_data = DNPData(data, dims, coords, attrs)

    return kea_data


# Depreciated
def import_prospa_dir(path, exp_list=None):
    """Import directory of prospa experiments

    Args:
        path (str): Directory of all data
        exp_list (list): List of directorys to include

    Returns:
        dict: Dictionary of DNPData objects
    """

    dirs = [x for x in os.listdir(path) if os.path.isdir(os.path.join(path, x))]

    if exp_list is not None:
        dirs = [dir_ for dir_ in dirs if dir_ in exp_list]

    ws = {}

    for ix, dir_ in enumerate(dirs):
        try:
            tmp = import_prospa(os.path.join(path, dir_))
            ws[dir] = tmp
        except ValueError as e:
            print("Skipping folder: %s" % str(e))

    return ws


def import_nd(path):
    """Import Kea binary 1d, 2d, 3d, 4d files

    Args:
        path (str): Path to file

    Returns:
        tuple:
            x (None, numpy.array): Axes if included in binary file, None otherwise
            data (numpy.array): Numpy array of data
    """

    ascii_header_size = 12
    ascii_header_format = "<4s4s4s"

    with open(path, "rb") as f:
        ascii_header_bytes = f.read(ascii_header_size)

        ascii_header = unpack(ascii_header_format, ascii_header_bytes)

        owner = ascii_header[0][::-1].decode("utf-8")
        format_ = ascii_header[1][::-1].decode("utf-8")
        version = ascii_header[2][::-1].decode("utf-8")

        # header size depends on version
        if version == "V1.0":
            header_format = "4i"
            header_size = 16
        else:
            header_format = "5i"
            header_size = 20

        header_bytes = f.read(header_size)

        header = unpack(header_format, header_bytes)

        dataType = header[0]
        xDim = header[1]
        yDim = header[2]
        zDim = header[3]
        if version != "V1.0":
            qDim = header[4]
        else:
            qDim = 1

        raw = f.read()

        x = None
        if dataType == 500:  # float
            raw_data = unpack("<%if" % (xDim * yDim * zDim * qDim), raw)
            data = np.array(raw_data)
        elif dataType == 501:  # complex
            raw_data = unpack("<%if" % (xDim * yDim * zDim * qDim * 2), raw)
            data = np.array(raw_data)
            data = data[0::2] - 1j * data[1::2]
        elif dataType == 502:  # double
            raw_data = unpack("<%id" % (xDim * yDim * zDim * qDim), raw)
            data = np.array(raw_data)
        elif dataType == 503:
            raw_data = unpack("<%if" % (xDim * yDim * zDim * qDim * 2), raw)
            raw_data = np.array(raw_data)
            x = raw_data[0:xDim]
            data = raw_data[xDim:]
        elif dataType == 504:
            raw_data = unpack("<%if" % (xDim * yDim * zDim * qDim * 3), raw)  # 504
            raw_data = np.array(raw_data)
            x = raw_data[0:xDim]
            data = raw_data[xDim:]
            data = data[0::2] - 1j * data[1::2]
        else:
            raise ValueError("Data %i type not recognized" % dataType)

        # re-shape using F-ordering (Fortran)
        data = data.reshape(xDim, yDim, zDim, qDim, order="F")

        data = data.squeeze()  # remove length 1 dimensions

    return x, data


def import_par(path):
    """Import Kea parameters .par file

    Args:
        path (str): Path to parameters file

    Returns:
        dict: Dictionary of Kea Parameters
    """

    attrs = {}

    with open(path, "r") as f:
        raw = f.read()
        lines = raw.rstrip().rsplit("\n")

        for line in lines:
            key, value = line.rstrip().rsplit(" = ")

            if value[0] == '"' and value[-1] == '"':
                value = value[1:-1]

            if "." in value:
                try:
                    value = float(value)
                except:
                    pass
            else:
                try:
                    value = int(value)
                except:
                    pass
            attrs[key] = value

    return attrs


def import_csv(path, return_raw=False, is_complex=True):
    """Import Kea csv file

    Args:
        path (str): Path to csv file

    Returns:
        tuple:
            x(numpy.array): axes if return_raw = False
            data(numpy.array): Data in csv file
    """

    raw = np.loadtxt(path, delimiter=",")

    if not return_raw:
        x = raw[:, 0]
        if is_complex:
            data = raw[:, 1::2] + 1j * raw[:, 2::2]
        else:
            data = raw[:, 1:]
        data = np.squeeze(data)
        return x, data
    else:
        return raw


def prospa_coords(attrs, data_shape, experiment):
    """Generate coords from prospa acquisition parameters

    Args:
        attrs (dict): Dictionary of prospa acqusition parameters
        data_shape (tuple): Shape of data

    Returns:
        tuple: dims and coords
    """

    if experiment is None:
        experiment = attrs["experiment"]
    dims = []
    coords = []

    if experiment == "1Pulse":
        pts = attrs["nrPnts"]
        dwell_time = attrs["dwellTime"]
        x = np.arange(0.0, pts * dwell_time, dwell_time) / 1e6
        dims.append("t2")
        coords.append(x)

    elif experiment == "B12T_1Pulse" or experiment == "B12T_1Pulse_MPS":
        pts = attrs["nrPnts"]
        dwell_time = attrs["dwellTime"]
        x = np.arange(0.0, pts * dwell_time, dwell_time) / 1e6
        dims.append("t2")
        coords.append(x)

        dims.append("Average")
        coords.append(np.arange(attrs["nrScans"]))

    elif experiment == "T1-IR-FID":
        pts = attrs["nrPnts"]
        dwell_time = attrs["dwellTime"]
        x = np.arange(0.0, pts * dwell_time, dwell_time) / 1e6
        dims.append("t2")
        coords.append(x)

        T1_steps = attrs["nrSteps"]
        T1_min_delay = attrs["minDelay"]
        T1_max_delay = attrs["maxDelay"]

        if attrs["delaySpacing"] == "lin":
            T1 = np.linspace(T1_min_delay, T1_max_delay, T1_steps) / 1000.0
        elif attrs["delaySpacing"] == "log":
            T1 = (
                np.logspace(np.log10(T1_min_delay), np.log10(T1_max_delay), T1_steps)
                / 1000.0
            )
        else:
            raise ValueError(
                f"Unable to determine delaySpacing {attrs['delaySpacing']}"
            )

        dims.append("t1")
        coords.append(T1)
    elif experiment == "B12T_T1-IR-FID" or experiment == "B12T_T1-IR-FID_MPS":
        pts = attrs["nrPnts"]
        dwell_time = attrs["dwellTime"]
        x = np.arange(0.0, pts * dwell_time, dwell_time) / 1e6
        dims.append("t2")
        coords.append(x)

        T1_steps = attrs["nrSteps"]
        T1_min_delay = attrs["minDelay"]
        T1_max_delay = attrs["maxDelay"]

        if attrs["delaySpacing"] == "lin":
            T1 = np.linspace(T1_min_delay, T1_max_delay, T1_steps) / 1000.0
        else:
            T1 = (
                np.logspace(np.log10(T1_min_delay), np.log10(T1_max_delay), T1_steps)
                / 1000.0
            )

        dims.append("t1")
        coords.append(T1)

        dims.append("Average")
        coords.append(np.arange(attrs["nrScans"]))
    elif experiment == "B12T_jres2D":
        pts = attrs["nrPnts"]
        dwell_time = attrs["dwellTime"]
        x = np.arange(0.0, pts * dwell_time, dwell_time) / 1e6
        dims.append("t2")
        coords.append(x)

        inter_pulse_delay = attrs["interPulseDelay"]
        increment = attrs["increment"]
        steps = attrs["nrSteps"]

        t1 = (
            np.r_[
                inter_pulse_delay : inter_pulse_delay
                + increment * (steps - 1) : 1j * steps
            ]
            / 1e6
        )

        dims.append("t1")
        coords.append(t1)

        dims.append("Average")
        coords.append(np.arange(attrs["nrScans"]))

    else:
        dims_list = ["x", "y", "z", "q"]
        for ix in range(len(data_shape)):
            dims.append(dims_list[ix])  # call dimensions in order: x, y, z, q
            coords.append(np.arange(data_shape[ix]))  # set coords to index for now
        if ("nrPnts" in attrs) and ("dwellTime" in attrs):
            pts = attrs["nrPnts"]
            dwell_time = attrs["dwellTime"]

            x = np.arange(0.0, pts * dwell_time, dwell_time) / 1e6

            dims[0] = "t2"
            coords[0] = x

    return dims, coords



import os
from .h5 import save_h5
from ..core.data import DNPData


def save(data_object, filename, save_type=None, *args, **kwargs):
    """Save data to h5 format

    Args:
        data_object (dnpdata object) : dnpdata object to save
        filename (str): name of file, must include extension .h5
        save_type (str): Type of file to save (optional). Allowed values: "h5"

    """

    if save_type == None:
        save_type = autodetect(filename)

    if save_type == "h5":
        if isinstance(data_object, dict):
            return save_h5(data_object, filename, *args, **kwargs)
        elif isinstance(data_object, DNPData):
            ws = {}
            ws["__DNPDATA__"] = data_object
            return save_h5(ws, filename, *args, **kwargs)
        else:
            raise TypeError(
                "object format not recognized, must be dnpdata or dnpdata_collection"
            )

    else:
        raise TypeError("File type not recognized, you must specify a save format")


def autodetect(test_name):

    if test_name[-1] == os.sep:
        test_name = test_name[:-1]

    if os.path.splitext(test_name)[1] == ".h5":
        type = "h5"
    else:
        raise TypeError("File type not recognized, you must specify a save format")

    return type



import numpy as np
from scipy.io import loadmat


def import_power(path, filename=""):
    """import powers file

    Args:
        path (str): Directory of powers
        filename (str): filename of powers if given

    Returns:
        t (numpy.ndarray): Array of time points
        p (numpy.ndarray): Array of powers
    """
    fullPath = path + filename

    if fullPath[-4:] == ".mat":
        rawDict = loadmat(fullPath)
        t = rawDict["timelist"].reshape(-1)
        p = rawDict["powerlist"].reshape(-1)

    elif fullPath[-4:] == ".csv":
        raw = np.loadtxt(fullPath, delimiter=",", skiprows=1)
        t = raw[:, 0].reshape(-1)
        p = raw[:, 1].reshape(-1)

    else:
        print("Could not identify power data type")
        return

    return t, p


def chop_power(t, p, threshold=0.1):
    """Use Derivative to chop Powers

    Args:
        t (numpy.ndarray): Array of time points
        p (numpy.ndarray): Array of powers
        threshold (float): Threshold to chop powers

    Returns:
        averageTimeArray: Array of average time values
        averagePowerArray: Array of average power values
    """

    diffPower = np.diff(p)

    step = [abs(x) > threshold for x in diffPower]

    correctedStep = []
    for ix in range(len(step) - 1):
        if step[ix] and step[ix + 1]:
            correctedStep.append(False)
        elif step[ix] and not step[ix + 1]:
            correctedStep.append(True)
        else:
            correctedStep.append(False)

    stepIndex = [0]
    for ix in range(len(correctedStep)):
        if correctedStep[ix]:
            stepIndex.append(ix)

    stepTupleList = []
    for ix in range(len(stepIndex) - 1):
        stepTupleList.append((stepIndex[ix], stepIndex[ix + 1]))

    averagePowerList = []
    averageTimeList = []
    for stepTuple in stepTupleList:
        averagePower = p[stepTuple[0] + 1 : stepTuple[1]]
        averagePower = np.mean(averagePower)
        averagePowerList.append(averagePower)
        averageTime = (t[stepTuple[0] + 1] + t[stepTuple[1]]) / 2.0
        averageTimeList.append(averageTime)

    averagePowerArray = np.array(averagePowerList)
    averageTimeArray = np.array(averageTimeList)
    return averageTimeArray, averagePowerArray


def assign_power(dataDict, expNumList, powersList):
    """Given a dictionary of dnpData objects with key being folder string,
    return the data with power values assigned to a new axis dimension

    Args:
        dataDict (dict): dictionary of data objects
        expNumList (list): List of experiment numbers
        powersList (list): List of powers

    Returns:
        DNPData: Data object with powers
    """

    doInitialize = True
    for ix, expNum in enumerate(expNumList):
        if str(expNum) in dataDict:
            if doInitialize:
                data = dataDict[str(expNum)]
                data.addAxes("power", powersList[ix])
                doInitialize = False
            else:
                tempData = dataDict[str(expNum)].copy()
                tempData.addAxes("power", powersList[ix])
                data.concatenateAlong(tempData, "power")

    return data



import unittest
import dnplab as dnp
from numpy.testing import assert_array_equal
import numpy as np


class dnpTools_tester(unittest.TestCase):
    def setUp(self):
        x = np.r_[0:10]
        y = x ** 2.0
        self.data = dnp.DNPData(y, ["t2"], [x])

    def test_signal_to_noise(self):

        dnp.signal_to_noise()

    def test_integrate(self):
        dnp.integrate(self.data, dim="t2")

    def test_mr_properties(self):

        info_1H = dnp.mr_properties("1H")
        self.assertEqual(info_1H, 26.7522128)

        info_1H = dnp.mr_properties("1H", 0.35)
        self.assertEqual(info_1H, 14902114.17018196)

        info_2H = dnp.mr_properties("2H", "qmom")
        self.assertEqual(info_2H, 0.286)

        info_6Li = dnp.mr_properties("6Li", "natAbundance")
        self.assertEqual(info_6Li, 7.59)

        info_6Li = dnp.mr_properties("6Li", "relSensitivity")
        self.assertEqual(info_6Li, 0.000645)



import numpy as np
from scipy import optimize
import warnings


def calculate_smax(spin_C=False):
    r"""Returns maximal saturation factor.

    Args:
        spin_C (float): unpaired spin concentration (M)

    Returns:
        smax (float): maximal saturation factor (unitless)

    .. math::
        \mathrm{s_{max}} = 1 - (2 / (3 + (3 * (\mathrm{spin\_C} * 198.7))))

    M.T. Türke, M. Bennati, Phys. Chem. Chem. Phys. 13 (2011) 3630. & J. Hyde, J. Chien, J. Freed, J. Chem. Phys. 48 (1968) 4211.
    """

    return 1 - (2 / (3 + (3 * (spin_C * 198.7))))


def interpolate_T1(
    E_powers=False,
    T1_powers=False,
    T1_array=False,
    interpolate_method="linear",
    delta_T1_water=False,
    T1_water=False,
    macro_C=False,
    spin_C=1,
    T10=2.0,
    T100=2.5,
):
    """Returns interpolated T1 data.

    Args:
        E_powers (numpy.array): The microwave powers at which to evaluate
        T1_powers (numpy.array): The microwave powers of the T1s to interpolate
        T1_array (numpy.array): The original T1s (s)
        interpolate_method (str): "second_order" or "linear"
        spin_C (float): unpaired electron spin concentration (M)
        T10 (float): T1 measured with unpaired electrons (s)
        T100 (float): T1 measured without unpaired electrons (s)
        delta_T1_water (optional) (float): change in T1 of water at max microwave power (s)
        T1_water (optional) (float): T1 of pure water (s)
        macro_C (optional) (float): concentration of macromolecule (M)

    Returns:
        interpolated_T1 (numpy.array): Array of T1 values same shape as E_powers and E_array

    T1 data is interpolated using Eq. 39 of http://dx.doi.org/10.1016/j.pnmrs.2013.06.001 for "linear" or Eq. 22 of https://doi.org/10.1016/bs.mie.2018.09.024 for "second_order"
    """

    # 2nd order fit, Franck and Han MIE (Eq. 22) and (Eq. 23)
    if interpolate_method == "second_order":
        if not macro_C:
            macro_C = spin_C

        if not delta_T1_water:
            delta_T1_water = T1_array[-1] - T1_array[0]
        if not T1_water:
            T1_water = T100

        kHH = (1.0 / T10 - 1.0 / T1_water) / macro_C
        krp = (
            (1.0 / T1_array)
            - (1.0 / (T1_water + delta_T1_water * T1_powers))
            - (kHH * (macro_C))
        ) / (spin_C)

        p = np.polyfit(T1_powers, krp, 2)
        T1_fit_2order = np.polyval(p, E_powers)

        interpolated_T1 = 1.0 / (
            ((spin_C) * T1_fit_2order)
            + (1.0 / (T1_water + delta_T1_water * E_powers))
            + (kHH * (macro_C))
        )

    # linear fit, Franck et al. PNMRS (Eq. 39)
    elif interpolate_method == "linear":

        linear_t1 = 1.0 / ((1.0 / T1_array) - (1.0 / T10) + (1.0 / T100))

        p = np.polyfit(T1_powers, linear_t1, 1)
        T1_fit_linear = np.polyval(p, E_powers)

        interpolated_T1 = T1_fit_linear / (
            1.0 + (T1_fit_linear / T10) - (T1_fit_linear / T100)
        )

    else:
        raise Exception("invalid interpolate_method")

    return interpolated_T1


def calculate_ksigma_array(powers=False, ksigma_smax=95.4, p_12=False):
    """Function to calcualte ksig array for any given ksigma and p_12

    Args:
        powers (numpy.array): Array of powers
        ksigma_smax (float): product of ksigma and smax (s^-1 * M^-1)
        p_12 (float): power at half max for ksigma fit

    Returns:
        ksig_fit (numpy.array): calculated ksigma array

    J.M. Franck et al. / Progress in Nuclear Magnetic Resonance Spectroscopy 74 (2013) 33–56
    """

    # Right side of Eq. 42. This function should fit to ksig_sp
    ksig_fit = (ksigma_smax * powers) / (p_12 + powers)

    return ksig_fit


def calculate_ksigma(ksigma_sp=False, powers=False, smax=1):
    """Get ksigma and E_power at half max of ksig

    Args:
        ksig (numpy.array): Array of ksigmas
        powers (numpy.array): Array of E_powers

    Returns:
        ksigma (float): calculated ksigma
        ksigma_stdd (float): standard deviation in ksigma
        p_12 (float): power at half max for ksigma fit

    J.M. Franck et al. / Progress in Nuclear Magnetic Resonance Spectroscopy 74 (2013) 33–56
    """

    # curve fitting
    # see https://docs.scipy.org/doc/scipy/reference/optimize.html
    popt, pcov = optimize.curve_fit(
        calculate_ksigma_array,
        powers,
        ksigma_sp,
        p0=[95.4 / 2, (max(powers) * 0.1)],
        method="lm",
    )

    assert popt[0] > 0, "Unexpected ksigma value: %d < 0" % popt[0]

    ksigma_smax = popt[0]
    p_12 = popt[1]
    ksigma_std = np.sqrt(np.diag(pcov))
    ksigma_stdd = ksigma_std[0] / smax

    ksigma_fit = calculate_ksigma_array(powers, ksigma_smax, p_12)

    ksigma = ksigma_smax / smax

    return ksigma, ksigma_stdd, ksigma_fit


def calculate_xi(tcorr=54e-12, omega_e=0.0614, omega_H=9.3231e-05):
    """Returns coupling_factor for any given tcorr

    Args:
        tcorr (float): translational diffusion correlation time (s)
        omega_e (float): electron gyromagnetic ratio
        omega_H (float): proton gyromagnetic ratio

    Returns:
        xi (float): coupling factor

    J.M. Franck et al. / Progress in Nuclear Magnetic Resonance Spectroscopy 74 (2013) 33–56
    """

    # Using Franck et al. PNMRS (2013)
    if tcorr < 0.1:
        tcorr *= 1e12

    zdiff = np.sqrt(1j * (omega_e - omega_H) * tcorr)
    zsum = np.sqrt(1j * (omega_e + omega_H) * tcorr)
    zH = np.sqrt(1j * omega_H * tcorr)

    # (Eq. 2)
    Jdiff = (1 + (zdiff / 4)) / (
        1 + zdiff + ((4 * (zdiff**2)) / 9) + ((zdiff**3) / 9)
    )

    Jsum = (1 + (zsum / 4)) / (1 + zsum + ((4 * (zsum**2)) / 9) + ((zsum**3) / 9))

    JH = (1 + (zH / 4)) / (1 + zH + ((4 * (zH**2)) / 9) + ((zH**3) / 9))

    # (Eq. 23) calculation of coupling_factor from the spectral density functions
    xi = ((6 * np.real(Jdiff)) - np.real(Jsum)) / (
        (6 * np.real(Jdiff)) + (3 * np.real(JH)) + np.real(Jsum)
    )

    return xi


def calculate_tcorr(coupling_factor=0.27, omega_e=0.0614, omega_H=9.3231e-05):
    """Returns translational correlation time (tcorr) in pico second

    Args:
        coupling_factor (float): coupling factor
        omega_e (float): electron gyromagnetic ratio
        omega_H (float): proton gyromagnetic ratio

    Returns:
        tcorr (float): translational diffusion correlation time (s)

    J.M. Franck et al. / Progress in Nuclear Magnetic Resonance Spectroscopy 74 (2013) 33–56
    """

    # root finding
    # see https://docs.scipy.org/doc/scipy/reference/optimize.html
    result = optimize.root_scalar(
        lambda t_corr: calculate_xi(t_corr, omega_e=omega_e, omega_H=omega_H)
        - coupling_factor,
        method="brentq",
        bracket=[1, 1e5],
    )

    if not result.converged:
        raise ValueError("Could not find tcorr")

    tcorr = result.root
    return tcorr * 1e-12


def calculate_uncorrected_Ep(
    uncorrected_xi=0.33,
    p_12_unc=0,
    E_powers=False,
    T10=2.0,
    T100=2.5,
    omega_ratio=658.5792,
    smax=1,
):
    """Function for E(p) for any given xi and p_12

    Args:
        uncorrected_xi (float): uncorrected coupling factor
        p_12_unc (float): power at half max for uncorrected_xi fit
        E_array (numpy.array): Array of enhancements
        E_powers (numpy.array): Array of E_powers
        T10 (float): T1(0), proton T1 with microwave power=0 (s)
        T100 (float): T10(0), proton T1 with spin_C=0 and microwave power=0 (s)
        omega_ratio (float): ratio of electron & proton gyromagnetic ratios
        smax (float): maximal saturation factor

    Returns:
        Ep_fit (numpy.array): uncorrected enhancement curve

    J.M. Franck et al. / Progress in Nuclear Magnetic Resonance Spectroscopy 74 (2013) 33–56
    """

    # Right side of Eq. 42. This function should fit to ksig_sp
    Ep_fit = 1 - (
        (uncorrected_xi * (1 - (T10 / T100)) * omega_ratio)
        * ((E_powers * smax) / (p_12_unc + E_powers))
    )

    return Ep_fit


def _residual_Ep(
    x,
    E_array: np.array,
    E_powers: np.array,
    T10: float,
    T100: float,
    omega_ratio: float,
    smax: float,
):
    """Function for residuals between E(p) for any given xi and p_12 and the experimental E_array

    Args:
        x (list): [uncorrected coupling factor, power at half max for uncorrected_xi fit]
        E_array (numpy.array): Array of enhancements
        E_powers (numpy.array): Array of E_power
        T10 (float): T1(0), proton T1 with microwave power=0 (s)
        T100 (float): T10(0), proton T1 with spin_C=0 and microwave power=0 (s)
        omega_ratio (float): ratio of electron & proton gyromagnetic ratios
        smax (float): maximal saturation factor

    Returns:
        Ep_fit (numpy.array): uncorrected enhancement curve

    J.M. Franck et al. / Progress in Nuclear Magnetic Resonance Spectroscopy 74 (2013) 33–56
    """

    return E_array - calculate_uncorrected_Ep(
        uncorrected_xi=x[0],
        p_12_unc=x[1],
        E_powers=E_powers,
        T10=T10,
        T100=T100,
        omega_ratio=omega_ratio,
        smax=smax,
    )


def calculate_uncorrected_xi(
    E_array=False,
    E_powers=False,
    T10=2.0,
    T100=2.5,
    omega_ratio=658.5792,
    smax=1,
):
    """Get coupling_factor and E_power at half saturation

    Args:
        E_array (numpy.array): Array of enhancements
        E_powers (numpy.array): Array of powers
        T10 (float): T1(0), proton T1 with microwave power=0 (s)
        T100 (float): T10(0), proton T1 with spin_C=0 and microwave power=0 (s)
        omega_ratio (float): ratio of electron & proton gyromagnetic ratios
        smax (float): maximal saturation factor

    Returns:
        uncorrected_xi (float): uncorrected coupling factor
        p_12_unc (float): power at half max for uncorrected_xi fit

    J.M. Franck et al.; Progress in Nuclear Magnetic Resonance Spectroscopy 74 (2013) 33–56
    """

    # least-squares fitting.
    # see https://docs.scipy.org/doc/scipy/reference/optimize.html
    results = optimize.least_squares(
        fun=_residual_Ep,
        x0=[0.27, (max(E_powers) * 0.1)],
        args=(E_array, E_powers, T10, T100, omega_ratio, smax),
        jac="2-point",
        method="lm",
    )
    if not results.success:
        raise ValueError("Could not fit Ep")
    assert results.x[0] > 0, "Unexpected coupling_factor value: %d < 0" % results.x[0]

    uncorrected_xi = results.x[0]
    p_12_unc = results.x[1]

    return uncorrected_xi, p_12_unc


def hydration(data={}, constants={}):
    """Function for performing ODNP calculations

    Args:
        data (dict)                   : keys and values are described in the example
        constants (dict)              : (optional) keys and values are described in the example

    Returns:
        (dict)                        : keys and values are described in the example

    J.M. Franck et al.; Progress in Nuclear Magnetic Resonance Spectroscopy 74 (2013) 33–56
    http://dx.doi.org/10.1016/j.pnmrs.2013.06.001

    J.M. Franck, S. Han; Methods in Enzymology, Chapter 5, Volume 615, (2019) 131-175
    https://doi.org/10.1016/bs.mie.2018.09.024
    """

    if not data:
        raise ValueError("Please supply a valid data dictionary, see example")

    if "hydration_inputs" in data.keys():
        warnings.warn(
            "The workspace concept is depreciated, see the example in the docs for the new syntax"
        )
        _data = data["hydration_inputs"]
        if "hydration_constants" in data.keys():
            constants = data["hydration_constants"]
        data = _data

    if "tcorr_bulk" in constants.keys() and constants["tcorr_bulk"] > 0.1:
        warnings.warn(
            "tcorr_bulk should be given in seconds, support for picoseconds will be removed in a future release"
        )
        constants["tcorr_bulk"] *= 1e-12

    if "macro_C" in constants.keys() and constants["macro_C"] > 0.1:
        warnings.warn(
            "macro_C should be given in molar, support for micromolar will be removed in a future release"
        )
        constants["macro_C"] *= 1e-6

    if data["spin_C"] > 0.1:
        warnings.warn(
            "spin_C should be given in molar, support for micromolar will be removed in a future release"
        )
        data["spin_C"] *= 1e-6

    if "field" in data.keys():
        warnings.warn(
            "keyword 'field' is depreciated, please use 'magnetic_field' from now on"
        )
        if "magnetic_field" in data.keys():
            warnings.warn(
                "you supplied both 'field' and 'magnetic_field', only 'magnetic_field' will be used"
            )
        else:
            data["magnetic_field"] = data["field"]
        data.pop("field")

    if data["magnetic_field"] > 3:
        warnings.warn(
            "magnetic_field should be given in T, support for mT will be removed in a future release"
        )
        data["magnetic_field"] *= 1e-3

    standard_constants = {
        "ksigma_bulk": 95.4,
        "krho_bulk": 353.4,
        "klow_bulk": 366,
        "tcorr_bulk": 54e-12,
        "D_H2O": 2.3e-9,
        "D_SL": 4.1e-10,
        "delta_T1_water": False,
        "T1_water": False,
        "macro_C": False,
    }
    # these constants have been compiled from the various ODNP literature

    odnp_constants = {**standard_constants, **constants}

    if data["smax_model"] == "tethered":
        # Option 1, tether spin label
        s_max = 1  # (section 2.2) maximal saturation factor

    elif data["smax_model"] == "free":
        # Option 2, free spin probe
        s_max = calculate_smax(data["spin_C"])  # from:
        # M.T. Türke, M. Bennati, Phys. Chem. Chem. Phys. 13 (2011) 3630. &
        # J. Hyde, J. Chien, J. Freed, J. Chem. Phys. 48 (1968) 4211.

    elif isinstance(data["smax_model"], float):
        # Option 3, manual input of smax
        if not (data["smax_model"] <= 1 and data["smax_model"] > 0):
            raise ValueError(
                "if given directly, smax must be type float between 0 and 1"
            )
        s_max = data["smax_model"]
    else:
        raise ValueError(
            "'smax_model' must be 'tethered', 'free', or a float between 0 and 1"
        )

    omega_e = 1.76085963023e-1 * data["magnetic_field"]
    # gamma_e in 1/ps for the tcorr unit, then correct by magnetic_field in T.
    # gamma_e is from NIST. The magnetic_field cancels in the following omega_ratio but you
    # need these individually for the spectral density functions later.

    omega_H = 2.6752218744e-4 * data["magnetic_field"]
    # gamma_H in 1/ps for the tcorr unit, then correct by magnetic_field in T.
    # gamma_H is from NIST. The magnetic_field cancels in the following omega_ratio but you
    # need these individually for the spectral density functions later.

    omega_ratio = (omega_e / (2 * np.pi)) / (omega_H / (2 * np.pi))
    # (Eq. 4-6) ratio of omega_e and omega_H, divide by (2*pi) to get angular
    # frequency units in order to correspond to S_0/I_0, this is also ~= to the
    # ratio of the resonance frequencies for the experiment, i.e. MW freq/RF freq

    if "T1_powers" in data.keys():
        T1p = interpolate_T1(
            E_powers=data["E_powers"],
            T1_powers=data["T1_powers"],
            T1_array=data["T1_array"],
            interpolate_method=data["interpolate_method"],
            delta_T1_water=odnp_constants["delta_T1_water"],
            T1_water=odnp_constants["T1_water"],
            macro_C=odnp_constants["macro_C"],
            spin_C=data["spin_C"],
            T10=data["T10"],
            T100=data["T100"],
        )
    else:
        if len(data["T1_array"]) == len(data["E_array"]):
            T1p = data["T1_array"]
        else:
            raise ValueError(
                "'T1_array' must be equal in length to 'E_array'. Otherwise give 'T1_powers' equal in length to 'T1_array' in order to interpolate."
            )

    ksigma_array = (1 - data["E_array"]) / (data["spin_C"] * omega_ratio * T1p)
    # (Eq. 41) this calculates the array of ksigma*s(p) from the enhancement array,
    # dividing by the T1 array for the "corrected" analysis

    ksigma, ksigma_stdd, ksigma_fit = calculate_ksigma(
        ksigma_array, data["E_powers"], s_max
    )
    # fit to the right side of Eq. 42 to get (ksigma*smax) and half of the E_power at s_max, called p_12 here

    krho = ((1 / data["T10"]) - (1 / data["T100"])) / (
        data["spin_C"]
    )  # (Eq. 36) "self" relaxivity, unit is s^-1 M^-1

    coupling_factor = ksigma / krho  # coupling factor, unitless

    tcorr = calculate_tcorr(coupling_factor, omega_e, omega_H)
    # (Eq. 21-23) this calls the fit to the spectral density functions. The fit
    # optimizes the value of tcorr in the calculation of coupling_factor, the correct tcorr
    # is the one for which the calculation of coupling_factor from the spectral density
    # functions matches the coupling_factor found experimentally. tcorr unit is ps

    Dlocal = (odnp_constants["tcorr_bulk"] / tcorr) * (
        odnp_constants["D_H2O"] + odnp_constants["D_SL"]
    )
    # (Eq. 19-20) local diffusivity, i.e. diffusivity of the water near the spin label

    klow = ((5 * krho) - (7 * ksigma)) / 3
    # section 6, (Eq. 13). this describes the relatively slowly diffusing water
    # near the spin label, sometimes called "bound" water.
    # This is defined in its most compact form in:
    # Frank, JM and Han, SI;  Chapter Five - Overhauser Dynamic Nuclear Polarization
    # for the Study of Hydration Dynamics, Explained. Methods in Enzymology, Volume 615, 2019
    # But also explained well in:
    # Franck, JM, et. al.; "Anomalously Rapid Hydration Water Diffusion Dynamics
    # Near DNA Surfaces" J. Am. Chem. Soc. 2015, 137, 12013−12023.

    xi_unc, p_12_unc = calculate_uncorrected_xi(
        data["E_array"],
        data["E_powers"],
        data["T10"],
        data["T100"],
        omega_ratio,
        s_max,
    )
    # (Eqs. 7 and 44) this calculates the coupling factor using the "uncorrected" analysis

    uncorrected_Ep = calculate_uncorrected_Ep(
        xi_unc,
        p_12_unc,
        data["E_powers"],
        data["T10"],
        data["T100"],
        omega_ratio,
        s_max,
    )
    # (Eqs. 7 and 44) this calculates the "uncorrected" enhancement array using xi_unc

    return {
        "uncorrected_Ep": uncorrected_Ep,
        "uncorrected_xi": xi_unc,
        "interpolated_T1": T1p,
        "ksigma_array": ksigma_array,
        "ksigma_fit": ksigma_fit,
        "ksigma": ksigma,
        "ksigma_stdd": ksigma_stdd,
        "ksigma_bulk_ratio": ksigma / odnp_constants["ksigma_bulk"],
        "krho": krho,
        "krho_bulk_ratio": krho / odnp_constants["krho_bulk"],
        "klow": klow,
        "klow_bulk_ratio": klow / odnp_constants["klow_bulk"],
        "coupling_factor": coupling_factor,
        "tcorr": tcorr,
        "tcorr_bulk_ratio": tcorr / odnp_constants["tcorr_bulk"],
        "Dlocal": Dlocal,
    }



import matplotlib.pyplot as plt
import numpy as np


def imshow(data, *args, **kwargs):  # TODO: drop unused args and kwargs
    """Image Plot for dnpdata object

    Args:
        data (dnpdata): dnpdata object for image plot
        args: args for matplotlib imshow function
        kwargs: kwargs for matplotlib imshow function

    Example::

       # Plotting a dnpdata object
       dnp.dnpResults.plt.figure()
       dnp.dnpResults.imshow(data)
       dnp.dnpResults.plt.show()

       # Plotting a workspace (dnpdata_collection)
       dnp.dnpResults.plt.figure()
       dnp.dnpResults.imshow(ws['proc'])
       dnp.dnpResults.plt.show()
    """

    dims = data.dims

    x_coord = data.coords[dims[1]]
    y_coord = data.coords[dims[0]]

    x_min = np.min(x_coord)
    x_max = np.max(x_coord)
    y_min = np.min(y_coord)
    y_max = np.max(y_coord)

    plt.imshow(data.values, aspect="auto", extent=[x_min, x_max, y_max, y_min])
    plt.xlabel(dims[1])
    plt.ylabel(dims[0])



import numpy as np
from ..core.data import DNPData


def ndalign(data, dim="f2", reference=None, center=None, width=None, average=None):
    """Alignment of NMR spectra using FFT Cross Correlation

    Args:
        all_data (object) : dnpdata object
        dim (str) : dimension to align along
        reference (numpy) : second dimension to align along
        center (float) : range center
        width (float) : range width

    Returns:
        dnpdata: Aligned data in container

    Examples:

        >>> import numpy as np
        >>> from matplotlib.pylab import *
        >>> import dnplab as dnp
        >>> x = np.r_[-10:10:100j]
        >>> y = np.r_[0,1,2]
        >>> values = dnp.math.lineshape.lorentzian(x.reshape(-1,1),y.reshape(1,-1),1)
        >>> data = dnp.DNPData(values, ['f2','sample'], [x, y])
        >>> figure('Before Alignment')
        >>> dnp.plot(data)
        >>> data_aligned = dnp.ndalign(data)
        >>> figure('After Alignment')
        >>> dnp.plot(data_aligned)
        >>> show()
    """

    proc_parameters = {"dim": dim}

    original_order = data.dims  # Original order of dims

    data.reorder([dim])  # Move dim to first dimension

    all_values = data.values  # Extract Data Values for alignment

    if center != None and width != None:
        start = center - 0.5 * width
        stop = center + 0.5 * width
    elif center == None and width == None:
        start = data.coords[dim][-1]
        stop = data.coords[dim][0]
    else:
        raise ValueError("selected range is not accpetable")

    values = data[dim, (start, stop)].values

    all_original_shape = all_values.shape
    original_shape = values.shape  # Preserve original shape

    all_align_dim_length = all_original_shape[0]
    align_dim_length = original_shape[0]  # length of dimension to align down

    all_values = all_values.reshape(all_align_dim_length, -1)
    values = values.reshape(align_dim_length, -1)  # Reshape to 2d

    new_shape = np.shape(values)

    dim2 = new_shape[1]

    abs_values = np.abs(values)

    if reference is None:
        reference = np.abs(values[:, -1])
    elif isinstance(reference, DNPData):
        reference = np.abs(reference.values)
        if average != None:
            reference = np.convolve(reference, np.ones(average), "same") / average

    ref_max_ix = np.argmax(reference)

    all_aligned_values = np.zeros_like(all_values)

    for ix in range(dim2):
        if average != None:
            abs_values[:, ix] = (
                np.convolve(abs_values[:, ix], np.ones(average), "same") / average
            )
        cor = np.correlate(
            abs_values[:, ix], reference, mode="same"
        )  # calculate cross-correlation
        max_ix = np.argmax(cor)  # Maximum of cross correlation
        delta_max_ix = max_ix - ref_max_ix  # Calculate how many points to shift
        all_aligned_values[:, ix] = np.roll(
            all_values[:, ix], -1 * delta_max_ix
        )  # shift values

    all_aligned_values = all_aligned_values.reshape(
        all_original_shape
    )  # reshape to original values shape

    data.values = all_aligned_values  # Add aligned values back to data object

    data.reorder(original_order)  # Back to original order

    proc_attr_name = "ndalign"
    data.add_proc_attrs(proc_attr_name, proc_parameters)

    return data


def align(data, dim="f2", dim2=None, center=None, width=None):
    """DEPRECIATED. Please use ndalign instead. Alignment of NMR spectra down given dimension or dimensions

    Args:
        all_data (object) : dnpdata object
        dim (str) : dimension to align along
        dim2 (str) : second dimension to align along
        center (float) : range center
        width (float) : range width

    Returns:
        dnpdata: Aligned data in container
    """

    if len(np.shape(data.values)) > 3:
        raise ValueError("Greater than 3-dimensional data is currently not supported")

    proc_parameters = {"dim": dim, "dim2": dim2}
    originalAxesOrder = data.dims

    if (dim2 is None) or (len(data.shape) == 2):
        if len(data.shape) > 2:
            raise ValueError(
                "data has more than 2 dimensions, 2nd dimension is ambiguous"
            )
        dim2 = None
        data.reorder([dim])
        dimIter = data.dims[-1]

    else:
        data.reorder([dim, dim2])
        dimIter = data.dims[-1]
    if center != None and width != None:
        start = center - 0.5 * width
        stop = center + 0.5 * width
    elif center == None and width == None:
        start = None
        stop = None
    else:
        raise ValueError("selected range is not accpetale")
    if dim2 == None:
        if start != None and stop != None:
            refData = data[dimIter, 0, dim, (start, stop)].values.reshape(-1)
        elif start == None and stop == None:
            refData = data[dimIter, 0].values.reshape(-1)
        else:
            raise ValueError("selected range is not accpetale")

        for ix in range(len(data.coords[dimIter])):
            tempData = data[dimIter, ix].values.reshape(-1)
            if start != None and stop != None:
                rangeData = data[dimIter, ix, dim, (start, stop)].values.reshape(-1)
            elif start == None and stop == None:
                rangeData = tempData
            else:
                raise ValueError("selected range is not accpetale")

            corrData = np.correlate(np.abs(rangeData), np.abs(refData), mode="same")
            shiftIx = np.argmax(corrData) - (
                len(corrData) / 2
            )  # subtract half length so spectrum is shifted relative to center, not edge
            shiftData = np.roll(tempData, -1 * int(np.round(shiftIx, 0)))
            data.values[:, ix] = shiftData
    else:

        for ix1 in range(len(data.coords[-1])):
            if start != None and stop != None:
                refData = data[dim, (start, stop)].values[:, 0, 0]
            elif start == None and stop == None:
                refData = data.values[:, 0, 0]
            else:
                raise ValueError("selected range is not accpetale")

            for ix2 in range(len(data.coords[dim2])):
                tempData = data.values[:, ix2, ix1]
                if start != None and stop != None:
                    rangeData = data[dim, (start, stop)].values[:, ix2, ix1]
                elif start == None and stop == None:
                    rangeData = tempData
                else:
                    raise ValueError("selected range is not accpetale")

                corrData = np.correlate(np.abs(rangeData), np.abs(refData), mode="same")
                shiftIx = np.argmax(corrData) - (
                    len(corrData) / 2
                )  # subtract half length so spectrum is shifted relative to center, not edge
                shiftData = np.roll(tempData, -1 * int(np.round(shiftIx, 0)))
                data.values[:, ix2, ix1] = shiftData

    data.reorder(originalAxesOrder)

    proc_attr_name = "align"
    data.add_proc_attrs(proc_attr_name, proc_parameters)

    return data



import numpy as np
import os
from .. import DNPData


def import_specman(path):
    """Import specman data and return DNPData object

    Args:
        path (str) : Path to either .d01 or .exp file

    Returns:
        specman_data (object) : DNPData object containing specman data
    """
    if path[-1] == os.sep:
        path = path[:-1]
    if path[-1] == "p":
        file_name_d01 = path.replace("exp", "d01")
        file_name_exp = path
    elif path[-1] == "1":
        file_name_exp = path.replace("d01", "exp")
        file_name_d01 = path
    else:
        raise TypeError("Incorrect file type, must be .d01 or .exp")

    params = load_specman_exp(file_name_exp)
    values, dims, coords, attrs = load_specman_d01(file_name_d01, params)

    specman_data = DNPData(values, dims, coords, attrs)

    return specman_data


def load_specman_exp(path):
    """Import parameter fields of specman data

    Args:
        path (str) : Path to either .d01 or .exp file

    Returns:
        params (dict) : dictionary of parameter fields and values
    """
    exp_file_opened = open(path, encoding="utf8", errors="ignore")
    file_contents = exp_file_opened.read().splitlines()
    exp_file_opened.close()
    params = {}
    c = ""
    for i in range(0, len(file_contents)):
        exp_content = str(file_contents[i])
        splt_exp_content = exp_content.split(" = ")
        if "[" in exp_content and "]" in exp_content and "=" not in exp_content:
            c = splt_exp_content[0].replace("[", "").replace("]", "")
        elif exp_content == "":
            c = "param"
        elif len(splt_exp_content) > 1:
            params[c + "_" + splt_exp_content[0]] = splt_exp_content[1]
        elif len(splt_exp_content) == 1 and exp_content != "":
            params[c + "_" + str(i)] = splt_exp_content
        else:
            pass

    return params


def load_specman_d01(path, params):
    """Import spectrum or spectra of specman data

    Args:
        path (str) : Path to either .d01 or .exp file
        params (dict) : dictionary of parameters from exp file

    Returns:
        abscissa (list) : coordinates of axes
        y_data (ndarray) : spectrum or spectra if >1D
        dims (list) : axes names
        params (dict) : updated parameters dictionary
    """

    if not params:
        params = {}

    file_opened = open(path, "rb")
    uint_read = np.fromfile(file_opened, dtype=np.uint32)
    file_opened.close()

    file_opened = open(path, "rb")
    float_read = np.fromfile(file_opened, dtype="<f4")
    file_opened.close()
    float_data_real = float_read[14 : uint_read[7] + 14]
    float_data_complex = float_read[uint_read[7] + 14 : len(float_read)]
    float_data_folded = float_data_real + 1j * float_data_complex

    if uint_read[2] == 1:
        y_data = np.reshape(float_data_folded, (uint_read[3]))
    elif uint_read[2] == 2:
        y_data = np.reshape(float_data_folded, (uint_read[4], uint_read[3]))
    elif uint_read[2] == 3:
        y_data = np.reshape(
            float_data_folded, (uint_read[5], uint_read[4], uint_read[3])
        )
    elif uint_read[2] == 4:
        y_data = np.reshape(
            float_data_folded, (uint_read[6], uint_read[5], uint_read[4], uint_read[3])
        )
    else:
        raise TypeError("DNPLab currently only supports up to 4D data")

    y_data = np.transpose(y_data)

    dims_full = ["t2", "t1", "t0", "t"]
    dims = dims_full[0 : uint_read[2]]
    axes_lengths = uint_read[9:13]

    abscissa = []
    for k in range(0, len(dims)):
        if dims[k] in params.keys():
            abscissa.append(params[dims[k]])
        else:
            abscissa.append(np.array(range(0, axes_lengths[k])))

    return y_data, dims, abscissa, params



"""Modules operating on a dnpData object to return a new dnpData object"""
from .align import *
from .apodization import *
from .fft import *
from .helpers import *
from .integration import *
from .offset import *
from .phase import *



import pytest
import operator
import unittest
from numpy.testing import assert_array_equal
from dnplab.core.base import ABCData
from dnplab.core.coord import Coords
import numpy as np
import random

test_dims = ["x", "y", "z", "p", "q", "r"]
num_random_tests = 10


def get_random_ABCData_list(seed_axis=0, seed_data=0):
    """
    Pre-generate a list of randomized tuple (ABCData, np.array) for all the
    test cases here
    """
    random.seed(seed_axis)
    random_axes = [
        random.sample(
            list(zip(test_dims, [np.r_[0 : random.randint(1, 6)] for _ in test_dims])),
            3,
        )
        for _ in range(3)
    ]
    random.seed(seed_data)
    ABCData_list = []
    for random_axis in random_axes:
        dims = [axis[0] for axis in random_axis]
        coords = [axis[1] for axis in random_axis]
        shape = [coord.size for coord in coords]
        values = np.random.randn(*shape)
        ABCData_list.append((ABCData(values, dims, coords), values))
    # ABCData_list.append((ABCData.ABCData_core(), np.array([])))  # UserWarning: Github #37
    return ABCData_list


random_ABCData_list = get_random_ABCData_list(seed_axis=0, seed_data=0)
random_ABCData_list_2 = get_random_ABCData_list(seed_axis=0, seed_data=1)


@pytest.mark.filterwarnings("ignore:divide by zero")
# See https://docs.python.org/3/library/operator.html#mapping-operators-to-functions
@pytest.mark.parametrize(
    "operator", [operator.add, operator.sub, operator.mul, operator.truediv]
)
@pytest.mark.parametrize("ABCData_value_tuple", random_ABCData_list)
@pytest.mark.parametrize("number", [-1.1j, -1.1, -1, 0, 1, 1.1, 1.1j])
def test_ABCData_core_math_operators_numeric(operator, ABCData_value_tuple, number):
    ABCData, values = ABCData_value_tuple
    assert_array_equal(operator(ABCData, number).values, operator(values, number))
    assert_array_equal(operator(number, ABCData).values, operator(number, values))


@pytest.mark.parametrize(
    "operator", [operator.add, operator.sub, operator.mul, operator.truediv]
)
@pytest.mark.parametrize("i_data", range(0, len(random_ABCData_list)))
def test_ABCData_core_math_operators(operator, i_data):
    ABCData, values = random_ABCData_list[i_data]
    ABCData_2, values_2 = random_ABCData_list_2[i_data]
    ABCData_3, values_3 = operator(ABCData, ABCData_2), operator(values, values_2)
    # assert self consistent
    assert ABCData_3._self_consistent()
    # assert values equal
    assert_array_equal(ABCData_3.values, values_3)


def test_ABCData_core_math_div_by_zero():
    with pytest.warns(RuntimeWarning):
        ABCData(
            values=np.array([1, 2, 3]), coords=[np.array([3, 2, 1])], dims=["x"]
        ) / 0


class dnplab_ABCData_core_tester(unittest.TestCase):
    def setUp(self):
        self.dims = test_dims
        random.sample(test_dims, random.randint(1, len(test_dims)))

    def construct_random_data(self):
        random_dims = random.sample(test_dims, random.randint(1, len(test_dims)))

        random_coords = [np.r_[0 : random.randint(1, 6)] for dim in random_dims]
        shape = [coord.size for coord in random_coords]

        random_values = np.random.randn(*shape)
        data = ABCData(random_values, random_dims, random_coords)
        return data, random_dims, random_values, random_coords

    def test_ABCData_core_init(self):
        for ix in range(num_random_tests):
            (
                data,
                random_dims,
                random_values,
                random_coords,
            ) = self.construct_random_data()
            self.assertTrue(data._self_consistent())
            assert_array_equal(data.values, random_values)
            for ix, dim in enumerate(random_dims):
                assert_array_equal(data.coords[dim], random_coords[ix])
            self.assertListEqual(data.dims, random_dims)

    def test_coord_type(self):
        data, _, _, _ = self.construct_random_data()
        self.assertEqual(type(data.coords), Coords)

    def test_ndim(self):
        values = np.r_[1:10].reshape(3, 3)
        x = np.r_[0:3]
        y = np.r_[0:3]

        data = ABCData(values, ["x", "y"], [x, y])

        self.assertEqual(data.ndim, 2)


class dnplab_ABCData_coord_tester(unittest.TestCase):
    def setUp(self):
        self.coord_inst_a = np.r_[0:10:1]
        self.coord_inst_b = np.r_[0:1:0.005]
        self.numpy_inst = np.r_[1:2:0.25]
        self.collection_inst = Coords(
            ["a", "b", "c"], [self.coord_inst_a, self.coord_inst_b, self.numpy_inst]
        )

    def test_get_str_uses_dim(self):
        assert_array_equal(self.collection_inst.dims, ["a", "b", "c"])

    def test_get_int_uses_index(self):
        assert_array_equal(self.collection_inst[0], self.coord_inst_a)

    def test_like_dict(self):
        assert_array_equal(self.collection_inst["a"], self.coord_inst_a)

    def test_reorder_only_1_dim(self):
        self.collection_inst.reorder(["c"])
        assert_array_equal(self.collection_inst.dims, ["c", "a", "b"])

    def test_reorder_all_3_dims(self):
        self.collection_inst.reorder(["b", "c", "a"])
        assert_array_equal(self.collection_inst.dims, ["b", "c", "a"])

    def test_rename(self):
        self.collection_inst.rename("a", "new_a")
        assert_array_equal(self.collection_inst.dims, ["new_a", "b", "c"])



from __future__ import division
import numpy as np
import operator
import functools
from collections import OrderedDict
from copy import deepcopy


class Coords(object):
    def __init__(self, dims, coords):
        """Object for storing axes information

        Attributes:
            dims (list): List of strings labeling dimensions
            coords (list): List of arrays for axes
        """
        # super(Coords, self).__init__()

        if not isinstance(dims, list):
            raise ValueError("dims must be a list")
        if not isinstance(coords, list):
            raise ValueError("coords must be a list")

        for index, coord in enumerate(coords):
            if isinstance(coord, (range, list)):
                coords[index] = np.array(coord)

        if self._check_dims:
            self._dims = dims
        else:
            raise TypeError("dims must be list of str")

        if self._check_coords:
            self._coords = coords
        else:
            raise TypeError("coords must be list of 1d numpy arrays")

    def _check_dims(self, dims):
        """Verify dims is a list of str

        Args:
            dims: Object to test for valid dims

        Returns:
            bool: True if valid dims, False otherwise
        """
        # Check if dims is list
        if not isinstance(dims, list):
            return False

        # Check if all members are str
        for dim in dims:
            if not isinstance(dim, str):
                return False

        # Check for duplicates
        if len(dims) != len(set(dims)):
            return False

        return True

    def _check_coords(self, coords):
        """Check if valid coords

        Args:
            coords: Object to test

        Returns:
            bool: True if list of 1d numpy arrays, False otherwise
        """

        # Verify coords is a list
        if not isinstance(coords, list):
            return False

        # Verify each member is 1d numpy array (and not empty)
        for coord in coords:
            if not isinstance(coord, np.ndarray):
                return False
            if (not (len(coord.shape) == 1)) or (coord.size == 0):
                return False

        return True

    def _self_consistent(self):
        return NotImplemented

    def index(self, dim):
        return self._dims.index(dim)

    def __getitem__(self, dim):
        if isinstance(dim, str):
            return self.coords[self.index(dim)]
        elif isinstance(dim, int):
            return self.coords[dim]
        else:
            raise TypeError("dim must be type str or int not: %s" % str(type(dim)))

    def __setitem__(self, dim, coord):
        if not isinstance(dim, str):
            raise TypeError("dim must be type str not %s" % str(type(dim)))
        if not isinstance(coord, (np.ndarray)):
            raise TypeError(
                "argument must be type nddata_coord or numpy ndarray not %s"
                % str(type(coord))
            )

        # if dim already in dims, overwrite
        if dim in self.dims:
            index = self.index(dim)
            self._coords[index] = coord
        else:
            self._dims.append(dim)
            self._coords.append(coord)

    def __delitem__(self, dim):
        index = self.index(dim)
        del self._dims[index]
        del self._coords[index]

    @property
    def dims(self):
        return self._dims

    @dims.setter
    def dims(self, dims):
        if self._check_dims(dims):
            self._dims = dims
        else:
            raise TypeError("Invalid dims. Cannot set dims to {}".format(dims))

    @property
    def coords(self):
        return self._coords

    @coords.setter
    def coords(self, coords):
        if self._check_coords(coords):
            self._coords = coords
        else:
            raise TypeError("Invalid coords. Cannot set coords to {}".format(coords))

    def __repr__(self):
        return "Coords({})".format(self.coords)

    def __str__(self):
        return "dims:\n{}\ncoords:\n{}".format(self.dims, self.coords)

    def __iter__(self):
        return iter(self.coords)

    def __len__(self):
        return len(self.coords)

    @property
    def shape(self):
        return tuple(k.size for k in self.coords)

    @property
    def size(self):
        return functools.reduce(operator.mul, [len(k) for k in self.coords], 1)

    def reorder(self, dims):

        if not self._check_dims(dims):
            raise TypeError("New dims must be list of str with no duplicates")
        for dim in dims:
            if dim not in self.dims:
                raise ValueError("no such dimension: %s" % dim)

        # Add original dims to end, remove duplicates
        dims = list(OrderedDict.fromkeys(dims + self.dims))

        # New indices for dims
        new_order = [dims.index(dim) for dim in self.dims]
        permutation_order = [self.dims.index(dim) for dim in dims]

        self.dims = dims

        #        self.coords = [self.coords[x] for x in new_order]
        self.coords = [self.coords[x] for x in permutation_order]

    def pop(self, dim):
        index = self.index(dim)
        out = self._coords.pop(index)
        self.dims.pop(index)
        return out

    def copy(self):
        return deepcopy(self)

    def reorder_index(self, new_order):
        """Reorder based on index"""
        self._coords = [self._coords[x] for x in new_order]
        self._dims = [self._dims[x] for x in new_order]

    def rename(self, dim, new_dim):

        self.dims[self.index(dim)] = new_dim

    def append(self, dim, coord):
        """Append to coords"""
        if not isinstance(dim, str):
            raise TypeError("dim must be type str not %s" % str(type(dim)))

        if not isinstance(coord, (complex, float, int, np.ndarray)):
            raise TypeError("coord must be type numpy not %s" % str(type(coord)))

        if isinstance(coord, (list, range, float, int, complex)):
            coord = np.array(coord)

        if dim not in self.dims:
            self._dims.append(dim)
            self._coords.append(coord)
        else:
            raise ValueError("dim already in dims, cannot append to coords")



"""
DNPData object for storing N-dimensional data with coordinates
"""

import numpy as np

from .base import ABCData
from ..version import __version__

version = __version__

core_attrs_list = ["nmr_frequency"]

np.set_printoptions(threshold=15)


class DNPData(ABCData):
    """
    DNPData Class for handling dnp data

    The DNPData class is inspired by pyspecdata nddata object which handles n-dimensional data, axes, and other relevant information together.

    This class is designed to handle data and axes together so that performing NMR processing can be performed easily.

    Attributes:
        values (numpy.ndarray): Numpy Array containing data
        coords (list): List of numpy arrays containing axes of data
        dims (list): List of axes labels for data
        attrs (dict): Dictionary of parameters for data


    """

    def __init__(self, values=np.r_[[]], dims=[], coords=[], attrs={}, proc_attrs=None):
        """
        DNPData Class __init__ method

        Args:
            data (numpy.ndarray):
            coords (list): list of axes
            dims (list): list of strings which are names of axes
            attrs (dict): dictionary of parameters
        """

        if len(dims) > 0 and isinstance(dims[0], list):
            dims = dims[0]

        super().__init__(values, dims, coords, attrs)
        self.version = version
        if proc_attrs is not None:
            self.proc_attrs = proc_attrs
        else:
            self.proc_attrs = []
        self.max_print_attrs = 5
        self.print_values = False

    @property
    def _constructor(self):
        return DNPData

    def __repr__(self):
        """
        Representation of DNPData object
        """
        return "nddata(values = {}, coords = {}, dims = {}, attrs = {})".format(
            repr(self.values), repr(self.coords), repr(self.dims), repr(self.attrs)
        )

    def __str__(self):
        """
        String representation of DNPData object
        """

        string = "values:\n\t"
        string += " x ".join(map(str, self.shape))

        string += " {} ({})\n".format(type(self.values).__name__, self.values.dtype)

        if self.print_values is True:
            string += str(self.values) + "\n"

        string += "dims:\n\t"

        string += "{}\n".format(self.dims)

        string += "coords:\n\t"
        string += "\n\t".join(map(repr, self.coords))

        string += "\n"

        string += "attrs:\n"

        for ix, key in enumerate(self.attrs.keys()):
            if ix == self.max_print_attrs:
                string += "\t+%i attrs" % (len(self.attrs) - self.max_print_attrs)
                break
            string += "\t{!r}: {!r}\n".format(key, self.attrs[key])

        return string

    def proc_info(self):
        """
        Print processing steps and parameters currently in proc_attrs list
        """

        print("-----------------")
        print("proccessing steps")
        print("-----------------")
        if not self.proc_attrs:
            print("none.")
        else:
            for x in self.proc_attrs:
                print(
                    x[0]
                    + ": "
                    + str([y + "=" + str(x[1][y]) for y in x[1].keys()])
                    .replace("[", "")
                    .replace("]", "")
                    .replace("'", "")
                )

    def add_proc_attrs(self, proc_attr_name, proc_dict):
        """
        Stamp processing step to DNPData object

        Args:
            proc_attr_name (str): Name of processing step (e.g. "fourier_transform")
            proc_dict (dict): Dictionary of processing parameters for this processing step.
        """
        if not isinstance(proc_attr_name, str):
            raise ValueError("Processing step name must be string")
        if not isinstance(proc_dict, dict):
            raise ValueError("Processing dictionary must be dictionary")

        self.proc_attrs.append((proc_attr_name, proc_dict))

    def phase(self):
        """
        Return phase of DNPData object

        Returns:
            phase (float,int): phase of data calculated from sum of imaginary
                divided by sum of real components
        """
        return np.arctan(np.sum(np.imag(self.values)) / np.sum(np.real(self.values)))

    def squeeze(self):
        """
        Remove all length 1 dimensions from data

        .. warning::
            Axes information is lost

        Example:
            data.squeeze()
        """
        remove_axes = []
        for axes_ix, axes_value in enumerate(self.coords):
            if len(axes_value) == 1:
                remove_axes.append(axes_ix)

        reverse_remove_axes = remove_axes[::-1]
        for index_ix, index_value in enumerate(reverse_remove_axes):
            self.coords.pop(index_value)
            self.dims.pop(index_value)
            self.values = np.squeeze(self.values)

    def select(self, selection):
        """
        Select subset of 2D data object

        Args:
            selection (int, range, list, tuple): list or tuple of slices to keep

        Returns:
            DNPData object: subset of DNPData object

        Example:
            data.select((1, range(5,10), 15)) # keeps slices: 1, 5, 6, 7, 8, 9, and 15

        """
        if len(self.dims) == 1:
            raise TypeError("select method is not applicable to 1D data")
        if isinstance(selection, int):
            self.values = self.values[:, selection]
            self.coords[self.dims[1]] = self.coords[self.dims[1]][selection]
            self.dims = self.dims[0]
        elif isinstance(selection, range):
            self.values = self.values[:, selection.start : selection.stop]
            self.coords[self.dims[1]] = self.coords[self.dims[1]][
                selection.start : selection.stop
            ]
        elif isinstance(selection, (list, tuple)) and all(
            [isinstance(x, (int, range)) for x in selection]
        ):
            new_values = np.empty(shape=(self.shape[0],))
            new_coords = []
            for x in selection:
                if isinstance(x, int):
                    new_values = np.vstack((new_values, self.values[:, x]))
                    new_coords = new_coords + [self.coords[self.dims[1]][x]]
                elif isinstance(x, range):
                    new_values = np.vstack(
                        (new_values, self.values[:, x.start : x.stop].T)
                    )
                    new_coords = new_coords + [
                        self.coords[self.dims[1]][y] for y in range(x.start, x.stop)
                    ]

            self.values = new_values.T[:, 1:]
            self.coords[self.dims[1]] = np.array(new_coords)
        else:
            raise TypeError(
                "Select using integer, range, or list/tuple of integers or ranges"
            )

        return self._constructor(
            values=self.values,
            coords=self.coords._coords,
            dims=self.dims,
            attrs=self.attrs,
            proc_attrs=self.proc_attrs,
        )



import numpy as np

import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button


def manual_phase(data, dim="f2"):
    """Manually Phase NMR Spectra"""

    fig, ax = plt.subplots()

    plt.title("Manual Phase")
    plt.subplots_adjust(bottom=0.25)
    init_phase = 0
    delta_phase = 1.0
    coord = data.coords[dim]
    size = coord.size
    values = data.values
    values = values.reshape(size, -1)
    values = values[:, 0].reshape(-1, 1)

    l0 = plt.plot(coord, values, alpha=0.5)
    l = plt.plot(coord, values)
    ax.margins(x=0)

    axcolor = "lightgoldenrodyellow"
    axphase = plt.axes([0.15, 0.1, 0.65, 0.03], facecolor=axcolor)
    axphase1 = plt.axes([0.15, 0.15, 0.65, 0.03], facecolor=axcolor)

    sphase = Slider(
        axphase, "Zeroth (deg)", -180, 180, valinit=init_phase, valstep=delta_phase
    )

    sphase1 = Slider(
        axphase1, "First (deg)", -180, 180, valinit=init_phase, valstep=delta_phase
    )

    def update(val):
        phase = sphase.val
        phase1 = sphase1.val
        margin = 0
        min_y_data = 0
        max_y_data = 0
        for ix, line in enumerate(l):
            y_data = np.real(
                np.exp(-1j * np.pi * phase / 180.0)
                * np.exp(-1j * np.pi * phase1 * coord / 180)
                * values[:, ix]
            )
            margin = max(0.1 * (np.max(y_data) - np.min(y_data)), margin)
            line.set_ydata(y_data)
            min_y_data = min(min_y_data, np.min(y_data))
            max_y_data = max(max_y_data, np.max(y_data))
        #        ax.set_ylim(min_y_data - margin, max_y_data + margin)
        fig.canvas.draw_idle()

    sphase.on_changed(update)
    sphase1.on_changed(update)

    resetax = plt.axes([0.8, 0.025, 0.1, 0.04])
    reset_button = Button(resetax, "Reset", color=axcolor, hovercolor="0.975")

    plus_90 = plt.axes([0.6, 0.025, 0.1, 0.04])
    plus_90_button = Button(plus_90, "+90", color=axcolor, hovercolor="0.975")

    minus_90 = plt.axes([0.4, 0.025, 0.1, 0.04])
    minus_90_button = Button(minus_90, "-90", color=axcolor, hovercolor="0.975")

    rescale_yaxis = plt.axes([0.2, 0.025, 0.1, 0.04])
    rescale_yaxis_button = Button(
        rescale_yaxis, "Rescale", color=axcolor, hovercolor="0.975"
    )

    def reset(event):
        sphase.reset()
        sphase1.reset()

    def plus_90_phase(event):
        new_phase = ((sphase.val + 270) % 360) - 180
        sphase.set_val(new_phase)

    def minus_90_phase(event):
        new_phase = ((sphase.val + 90) % 360) - 180
        sphase.set_val(new_phase)

    def rescale(event):
        ax.relim()
        ax.autoscale_view()

    reset_button.on_clicked(reset)
    plus_90_button.on_clicked(plus_90_phase)
    minus_90_button.on_clicked(minus_90_phase)
    rescale_yaxis_button.on_clicked(rescale)

    plt.show()
    manual_phase = sphase.val
    manual_phase1 = sphase1.val

    data *= np.exp(-1j * np.pi * manual_phase / 180.0)

    proc_attr_name = "manualphase"
    proc_parameters = {"phase": manual_phase, "phase1": manual_phase1}
    data.add_proc_attrs(proc_attr_name, proc_parameters)
    data.attrs["phase0"] = manual_phase
    data.attrs["phase1"] = manual_phase1

    return data



# TODO: remove unused imports
import numpy as np
import os
import re
from matplotlib.pylab import *

from .. import DNPData


def import_vna(path):
    """Import VNA data and return dnpdata object"""
    x, data = import_snp(path)
    # Not General
    dnpDataObject = DNPData(data, [x], ["f"], {})

    return dnpDataObject


# TODO: remove prints or make them optional
def import_snp(path):
    """Import sNp file and return numpy array"""
    path_filename, extension = os.path.splitext(path)

    extension_reg_ex = "[.]s[0-9]{1,}p"
    print(re.fullmatch(extension_reg_ex, extension))
    print(extension)
    if re.fullmatch(extension_reg_ex, extension) == None:
        raise ValueError("File Extension Not Given, Unspecified sNp file")

    num_reg_ex = "[0-9]{1,}"
    num = int(re.search(num_reg_ex, extension)[0])

    print(num)
    if num > 2:
        raise ValueError("Currently on s1p and s2p files are supported")

    f = open(path)
    read_string = " "
    while read_string[0] != "#":
        read_string = f.readline()
    raw = np.genfromtxt(f, skip_header=2, defaultfmt="11f")
    f.close()

    if num == 1:
        x = raw[:, 0]
        data = raw[:, 1] + 1j * raw[:, 2]

    if num == 2:
        x = raw[:, 1]

        data = np.zeros((len(x), 2, 2))

        data[:, 0, 0] = raw[:, 1] + 1j * raw[:, 2]  # S11
        data[:, 1, 0] = raw[:, 3] + 1j * raw[:, 4]  # S21
        data[:, 0, 1] = raw[:, 5] + 1j * raw[:, 6]  # S12
        data[:, 1, 1] = raw[:, 7] + 1j * raw[:, 8]  # S22

    if num > 2:
        x = raw[0::num]
        data = np.zeros((len(x), num, num))

        # TODO: Use list comprehension instead of two for loops
        for n in range(num):

            for m in range(num):

                data[:, n, m] = raw[n::num, 1 + 2 * m] + 1j * raw[n::num, 2 * (1 + m)]

    return x, data



import matplotlib.pyplot as plt
import numpy as np


def stack(data, *args, offset=None, **kwargs):
    """Stack Plot for 2D data

    Args:
        data (dnpdata): dnpdata object for matplotlib plot function
        args: args for matplotlib plot function
        offset: Value to offset each spectra, by default maximum of absolute value
        kwargs: kwargs for matplotlib plot function

    Example::

       dnp.dnpResults.plt.figure()
       dnp.dnpResults.stack(data)
       dnp.dnpResults.plt.show()

    """

    coord = data.coords[0]
    dim = data.dims[0]

    if offset == None:
        offset = np.max(data.abs)

    offset_matrix = (
        offset * np.ones(coord.size).reshape(-1, 1) * np.r_[0 : data.coords[1].size]
    )

    plt.plot(coord, data.values + offset_matrix, *args, **kwargs)


def waterfall(data, dx, dy, *args, **kwargs):
    """Waterfall plot for 2d data

    Args:
        data (dnpData): 2d Data object for waterfall plot
        dx (float, int): x-increment for each line
        dy (float, int): y-increment for each line

    Example::

       dnp.dnpResults.plt.figure()
       dnp.dnpResults.waterfall(data)
       dnp.dnpResults.plt.show()

    """

    coord = data.coords[0]
    dim = data.dims[0]

    for ix in range(data.coords[data.dims[1]].size):
        plt.plot(
            coord + (ix * dx),
            data[data.dims[1], ix].values.ravel() + (ix * dy),
            *args,
            **kwargs,
            zorder=-1 * ix
        )
        plt.fill_between(
            coord + (ix * dx),
            data[data.dims[1], ix].values.ravel() + (ix * dy),
            ix * dy,
            facecolor="w",
            edgecolor="None",
            zorder=-1 * ix,
        )



import numpy as np


def t1(t, T1, M_0, M_inf):
    """Exponential recovery for inversion recovery and saturation recovery T1 Measurements

    Args:
        t (array_like): time series
        T_1 (float): T1 value
        M_0 (float): see equation
        M_inf (float): see equation

    Returns:
        ndarray: T1 curve

    .. math::
        f(t) = M_{\infty} - (M_{\infty} - M_0) e^{-t/T_1}
    """

    return M_inf - (M_inf - M_0) * np.exp(-1.0 * t / T1)


def t2(t, M_0, T2, p=1.0):
    """Calculate stretched or un-stretched (p=1) exponential T2 curve

    Args:
        t (array_like): time series
        M_0 (float): see equation
        T_2 (float): T2 value
        p (float): see equation

    Returns:
        ndarray: T2 curve

    .. math::
        f(t) = M_{0} e^{(-(t/T_{2})^{p}}
    """

    return M_0 * np.exp(-1.0 * (t / T2) ** p)


def general_exp(t, C1, C2, tau):
    """Calculate mono-exponential curve

    Args:
        t (array_like): time series
        C1 (float): see equation
        C2 (float): see equation
        tau (float): see equation

    Returns:
        ndarray: mono-exponential curve

    .. math::
        f(t) = C1 + C2 e^{-t/tau}
    """

    return C1 + C2 * np.exp(-1.0 * t / tau)


def general_biexp(t, C1, C2, tau1, C3, tau2):
    """Calculate bi-exponential curve

    Args:
        t (array_like): time series
        C1 (float): see equation
        C2 (float): see equation
        C3 (float): see equation
        tau1 (float): see equation
        tau2 (float): see equation

    Returns:
        ndarray: bi-exponential curve

    .. math::
        f(t) = C1 + C2 e^{-t/tau1} + C3 e^{-t/tau2}
    """

    return C1 + C2 * np.exp(-1.0 * t / tau1) + C3 * np.exp(-1.0 * t / tau2)


def buildup_function(p, E_max, p_half):
    """Calculate asymptotic buildup curve

    Args:
        p (array): power series
        E_max (float): maximum enhancement
        p_half (float): power at half saturation

    Returns:
        ndarray: buildup curve

    .. math::
        f(p) = 1 + E_{max} * p / (p_{1/2} + p)
    """

    return 1 + E_max * p / (p_half + p)


def ksigma_smax(p, E_max, p_half):
    """Calculate asymptotic buildup curve

    Args:
        p (array): power series
        E_max (float): maximum enhancement
        p_half (float): power at half saturation

    Returns:
        ndarray: buildup curve

    .. math::
        f(p) = E_{max} * p / (p_{1/2} + p)
    """

    return E_max * p / (p_half + p)


def logistic(x, c, x0, L, k):
    """Not Implemented. Placeholder for calculating asymptotic buildup curve

    Args:
        x (array): x values
        c (float): offset
        x0 (float): x-value of sigmoid's midpoint
        L (float): maximum value
        k (float): logistic growth steepness

    Returns:
        ndarray: buildup curve
    """

    # return c + L / (1.0 + np.exp(-1.0 * k * (x - x0)))
    return NotImplemented



from .data import DNPData
import numpy as np


def concat(data_list, dim, coord=None):
    """Concatenates list of data objects down another dimension

    args:
        data_list (list): List of DNPData objects to concatentate
        dim (str): new dimension name
        coord: coords for new dimension

    Returns:
        data (DNPData): concatenated data object

    """

    shape = data_list[0].shape
    values_list = [data.values for data in data_list]

    for values in values_list:
        this_shape = values.shape
        if this_shape != shape:
            raise IndexError(
                "Cannot concatenate data objects. Array shapes do not match.",
                this_shape,
                shape,
            )

    dims = data_list[0].dims
    coords = data_list[0].coords.coords
    attrs = data_list[0].attrs

    values = np.stack(values_list, axis=-1)

    dims.append(dim)

    if coord is None:
        coords.append(np.arange(len(data_list)))
    else:
        coords.append(coord)

    data = DNPData(values, dims, coords, attrs)

    return data



import numpy as np
import os

from .. import DNPData

from struct import unpack

headerSize = 32
blockHeaderSize = 28

header_fmt = ">llllllhhl"
blockHeader_fmt = ">hhhhlffff"


def array_coords(attrs):
    """Return array dimension coords from parameters dictionary

    Args:
        attrs (dict): Dictionary of procpar parameters

    Returns:
        tuple: dim and coord for array
    """

    try:

        array_delta = attrs["arraydelta"]
        array_dim = attrs["arraydim"]
        array_start = attrs["arraystart"]
        array_stop = attrs["arraystop"]

        if array_dim != 1:
            coord = np.r_[array_start : array_stop + array_delta : array_delta]
            dim = "t1"
        else:
            coord = None
            dim = None

        return dim, coord

    except KeyError:

        coord = None
        dim = None

        return dim, coord


def import_fid(path, filename="fid"):
    """Import VnmrJ fid file

    Args:
        path (str): Directory of fid file
        filename (str): Name of fid file. "fid" by default

    Returns:
        numpy.ndarray: Array of data

    """
    with open(os.path.join(path, filename), "rb") as f:
        headerString = f.read(headerSize)
        header = unpack(header_fmt, headerString)

        nblocks = header[0]  # number of blocks in file
        ntraces = header[1]  # number of traces per block
        npts = header[2]  # number of elements per trace
        ebytes = header[3]  # number of bytes per element
        tbytes = header[4]  # number of bytes per traces
        bbytes = header[5]  # number of bytes per block
        vers_id = header[6]  # software version, file id status bits
        status = header[7]  # status of whole file
        nbheaders = header[8]  # number of block headers per block

        # check if int or float
        isFloat = False
        if status & 0x08:
            isFloat = True

        dataList = []
        for ix in range(nblocks):
            blockHeaderString = f.read(blockHeaderSize)
            blockHeader = unpack(blockHeader_fmt, blockHeaderString)

            scale = blockHeader[0]  # scaling factor
            block_status = blockHeader[1]  # status of data in block
            index = blockHeader[2]  # block index
            mode = blockHeader[3]  # mode of data in block
            ctcount = blockHeader[4]  # ct value for FID
            lpval = blockHeader[5]  # f2 (2D-f1) left phase in phasefile
            rpval = blockHeader[6]  # f2 (2D-f1) right phase in phasefile
            lvl = blockHeader[7]  # level drift correction
            tlt = blockHeader[8]  # tilt drift correction

            blockDataString = f.read(tbytes)

            if isFloat:
                blockData = np.array(
                    unpack(">%if" % (npts), blockDataString), dtype=complex
                )

            else:
                blockData = np.array(unpack(">%ii" % (npts), blockDataString))
            data = blockData[0::2] - 1j * blockData[1::2]  # minus sign for VNMRJ data
            dataList.append(data)
        dataArray = np.array(dataList).T

    return dataArray


def import_procpar(path, filename="procpar"):
    """Import VnmrJ procpar parameters file

    Args:
        path (str): Directory of file

    Returns:
        dict: Dictionary of procpar parameters
    """
    paramDict = {}
    with open(os.path.join(path, filename), "r") as f:
        while True:
            line = f.readline()
            if line == "":
                return paramDict
            else:
                # Line 1: Name & Type Line
                splitLine = line.rstrip().split(" ")

                name = splitLine[0]
                subtype = splitLine[1]
                basictype = int(splitLine[2])
                maxvalue = float(splitLine[3])
                minvalue = float(splitLine[4])
                stepsize = float(splitLine[5])
                Ggroup = int(splitLine[6])
                Dgroup = int(splitLine[7])
                protection = int(splitLine[8])
                active = int(splitLine[9])
                intptr = int(splitLine[10])

                # 3 Cases:
                # basictype is 1 (real) -> Line 2 separated by spaces
                # basictype is 2 (string) & First number is 1 -> single string on same line inside double quotes
                # basic type is 2 (string) & first number is greater than 1 -> first element is on same line, subsequent elements are on next lines, strings are surrounded by double quotes

                # Line 2: Value line
                firstValueLine = f.readline()
                valueLine = firstValueLine.rstrip().split(" ")
                numValues = int(valueLine[0])

                if basictype == 1:
                    if numValues == 1:
                        value = float(valueLine[1])
                    else:
                        listFloats = []
                        for number in valueLine[1:]:
                            listFloats.append(float(number))

                        value = listFloats

                elif basictype == 2:
                    if numValues == 1:
                        value = valueLine[1].replace('"', "")
                    else:
                        listStrings = []
                        listStrings.append(valueLine[1].replace('"', ""))

                        for ix in range(numValues - 1):
                            nextValueLine = f.readline()
                            nextValue = nextValueLine.strip()
                            listStrings.append(nextValue.replace('"', ""))

                        value = listStrings

                finalLine = f.readline()
                enumValuesLine = finalLine.rstrip().split(" ")
                numEnumValues = enumValuesLine[0]

                if int(numEnumValues) == 1:
                    enumValues = enumValuesLine[1]
                else:
                    enumValues = enumValuesLine[1:]

                paramDict[name] = value


def import_vnmrj(path, fidFilename="fid", paramFilename="procpar"):
    """Import VnmrJ Data

    Args:
        path(str): path to experiment folder
        fidFilename(str): FID file name
        paramFilename(str): process parameter filename

    Returns:
        dnpdata: data in dnpdata object
    """

    attrs = import_procpar(path, paramFilename)

    nmr_frequency = attrs["H1reffrq"] * 1.0e6
    sw = attrs["sw"]
    npts = int(attrs["np"] / 2)

    attrs["nmr_frequency"] = nmr_frequency

    dim, coord = array_coords(attrs)

    dwellTime = 1.0 / sw

    t = np.r_[0.0 : int(npts)] * dwellTime
    dims = ["t2"]
    coords = [t]

    data = import_fid(path, fidFilename)

    if coord is not None:
        dims.append(dim)
        coords.append(coord)
    else:
        data = data.reshape(-1)

    return DNPData(data, dims, coords, attrs)



from .. import DNPData
import numpy as np
import struct
import re


def import_tnmr(path):
    """Import tnmr data and return DNPData object

    Args:
        path (str) : Path to .jdf file

    Returns:
        tnmr_data (object) : DNPData object containing tnmr data
    """

    attrs = import_tnmr_pars(path)
    values, dims, coords = import_tnmr_data(path)

    tnmr_data = DNPData(values, dims, coords, attrs)

    return tnmr_data


def import_tnmr_pars(path):
    """Import parameter fields of tnmr data

    Args:
        path (str) : Path to .tnt file

    Returns:
        params (dict) : dictionary of parameter fields and values
    """

    params = {}

    with open(path, "rb") as f:
        params["version"] = f.read(8).decode("utf-8")

    return params


def import_tnmr_data(path):
    """Import spectrum or spectra of tnmr data

    Args:
        path (str) : Path to .tnt file

    Returns:
        data (ndarray) : spectrum or spectra if >1D
        abscissa (list) : coordinates of axes
        dims (list) : axes names
    """

    with open(path, "rb") as f:

        version = f.read(8).decode("utf-8")

        section = None

        while section != "":

            section = f.read(4).decode("utf-8")
            section = str(section)

            if section == "TMAG":
                flag = bool(f.read(4))
                if flag:
                    bytes_to_read = f.read(4)
                    bytes_to_read = struct.unpack("<i", bytes_to_read)[0]

                    header = f.read(bytes_to_read)

                    ### Deal With Header Here ###

            elif section == "DATA":
                flag = bool(f.read(4))
                if flag:
                    bytes_to_read = f.read(4)
                    bytes_to_read = struct.unpack("<i", bytes_to_read)[0]

                    raw_data = f.read(bytes_to_read)

                    raw_data = struct.unpack("%if" % (bytes_to_read / 4), raw_data)

                    raw_data = np.array(raw_data)

                    data = raw_data[::2] + 1j * raw_data[1::2]

            else:
                flag = bool(f.read(4))
                if flag:
                    bytes_to_read = f.read(4)
                    bytes_to_read = struct.unpack("<i", bytes_to_read)[0]

                    unsupported_bytes = f.read(bytes_to_read)

    abscissa = np.array(range(0, len(data)))

    dims = ["t2"]
    coords = [abscissa]

    return data, dims, coords



from __future__ import division
import operator
import numpy as np
import warnings
from copy import deepcopy
from collections import OrderedDict
from .coord import Coords

from ..version import __version__

version = __version__


_numerical_types = (np.ndarray, int, float, complex, np.complex64)

# _nddata_core_version = "1.0"


class ABCData(object):
    """N-Dimensional Data Object

    Attributes:
        values (numpy.ndarray): Data values in
        dims (list): List of strings giving dimension labels
        coords (Coords): Collection of numpy.ndarrays defining the axes
        attrs (dict): dictionary of parameters
        error (numpy.ndarray): If not None, error for values which are propagated during mathematical operations
        proc_attrs (list): List of processing steps
    """

    __array_priority__ = 1000  # radd, rsub, ... should return nddata object

    def __init__(
        self, values=np.r_[[]], dims=[], coords=[], attrs={}, error=None, **kwargs
    ):

        self.version = version

        # if values is list, convert to numpy array
        if isinstance(values, list):
            values = np.array(values)

        # verify values are numpy array
        if isinstance(values, np.ndarray):
            self._values = values
        else:
            raise TypeError(
                'values must be type "numpy.ndarray" not %s' % str(type(values))
            )

        self._coords = Coords(dims, coords)

        if isinstance(attrs, dict):
            self._attrs = attrs
        else:
            raise TypeError('attrs must be type "dict" not %s' % str(type(attrs)))

        if isinstance(error, np.ndarray) or (error == None):
            self._error = error
        else:
            raise TypeError(
                'error must be type "numpy.ndarray" or "None" not %s' % str(type(error))
            )

        if "proc_attrs" in kwargs:
            proc_attrs = kwargs["proc_attrs"]
            if isinstance(proc_attrs, list):
                self._proc_attrs = proc_attrs
            else:
                raise TypeError

        if not self._self_consistent():
            warnings.warn("Dimensions not consistent")

    @property
    def __version__(self):
        return self._nddata_core_version

    def _check_dims(self, dims):
        """Check that list is a list of strings with no duplicates

        Args:
            dims list to verify as list of strings

        Returns:
            bool: True if arguments is list of strings with no duplicates, False otherwise
        """

        # test that all members are strings
        all_strings = all([isinstance(dims[x], str) for x in range(len(dims))])

        # test if any duplicates exist
        any_duplicates = len(dims) == len(set(dims))

        return all_strings and any_duplicates

    def _check_coords(self, coords):
        """Check that coords is list of 1d numpy arrays

        Args:
            coords: coords object to check if valid coords

        Returns:
            bool: True if valid coords. False, otherwise.
        """

        for coord in coords:
            if isinstance(coord, np.ndarray):
                if len(coord.shape) == 1:
                    pass
                else:
                    return False
            else:
                return False

        return True

    def _check_error(self, error):
        """Check that error attribute is numpy array and it's size matches values

        Args:
            error: Object to check if valid error

        Returns:
            bool: True if valid error. False otherwise.
        """

        check_type = isinstance(error, np.ndarray)

        if check_type:
            check_size = error.size == self._values.size
        else:
            check_size = False

        return check_type and check_size

    def _self_consistent(self):
        """Check if dimensions for values, dims, and coords are consistent

        Returns:
            bool: True if consistent, False otherwise
        """

        if self._values.size == 0:
            coords_check = len(self._coords) == 0
        else:
            coords_check = list(self._values.shape) == list(self.coords.shape)

        dims_check = len(self.values.shape) == len(self.dims)

        return coords_check and dims_check

    def _attrs_valid(self):
        """Verify attrs attribute is valid. All values in attrs must be list, numpy.ndarray, int, float, complex, or str.

        Returns:
            bool: True if attrs is valid. False, otherwise.
        """

        for key in self._attrs:
            if not isinstance(
                self._attrs, (list, np.ndarray, int, float, complex, str)
            ):
                return False

        return True

    def __getitem__(self, args):
        """Method for indexing nddata_core

        Args:
            args (tuple): Tuple containing alternative dims and indexing values for each dimension to be indexed. (e.g. data['x', 1:10, 'y', :, 'z', (3.5, 7.5)])

        Example::

            data['x', 1] # return data indexing down "x" dim with index 1

            data['x', 4.5] # return data indexing down "x" dim with "x" coords nearest to 4.5

            data['x', 2:5] # return data indexing down "x" dim with index from 2 to 5

            data['x', (100., 150.)] # return data indexing down "x" dim where coords are range from 100 to 150
        """
        a = self.copy()
        if len(args) % 2 == 1:
            raise ValueError("Cannot index with odd number of arguments")

        index_dims = args[0::2]
        for dim in index_dims:
            if dim not in a.dims:
                raise ValueError("dim not in dims")

        index_slice = args[1::2]

        # check slices
        for slice_ in index_slice:

            # type must be slice or tuple
            if not isinstance(slice_, (slice, tuple, float, int)):
                raise ValueError("Invalid slice type")

            # if tuple, length must be two: (start, stop)
            if isinstance(slice_, tuple) and not len(slice_) in (1, 2):
                raise ValueError("tuple index must have one or two values")

        # convert tuple to slice
        updated_index_slice = []
        for dim, slice_ in zip(index_dims, index_slice):
            if isinstance(slice_, tuple):
                index = a.index(dim)
                if len(slice_) == 1:
                    start = np.argmin(np.abs(slice_[0] - a.get_coord(dim)))
                    updated_index_slice.append(slice(start, start + 1))
                else:
                    start = np.argmin(np.abs(slice_[0] - a.get_coord(dim)))
                    stop = np.argmin(np.abs(slice_[1] - a.get_coord(dim)))
                    if start == stop:
                        stop = start + 1
                    if stop < start:
                        start, stop = stop, start
                    updated_index_slice.append(slice(start, stop))
            elif isinstance(slice_, int):
                start = slice_
                if slice_ != -1:
                    updated_index_slice.append(slice(start, start + 1))
                else:
                    updated_index_slice.append(slice(slice_, None))
            elif isinstance(slice_, float):
                start = np.argmin(np.abs(slice_ - a.get_coord(dim)))
                updated_index_slice.append(slice(start, start + 1))
            else:
                updated_index_slice.append(slice_)

        index_slice_dict = dict(zip(index_dims, updated_index_slice))
        new_slices = [
            slice(None) if dim not in index_dims else index_slice_dict[dim]
            for dim in a.dims
        ]

        for ix, dim in enumerate(a.dims):
            a.coords[dim] = a.coords[dim][new_slices[ix]]

        a.values = a.values[tuple(new_slices)]

        return a

    def __setitem__(self, args, new_values):
        """Method for setting values by index nddata_core

        Args:
            args (tuple): Tuple containing alternative dims and indexing values for each dimension to be indexed. (e.g. data['x', 1:10, 'y', :, 'z', (3.5, 7.5)])
            new_values (numpy.ndarray): New values at given index

        Example::

            data['x', 1] = 1 # return data indexing down "x" dim with index 1

        """
        a = self.copy()

        if isinstance(new_values, ABCData):
            new_values = new_values.values

        if len(args) % 2 == 1:
            raise ValueError("Cannot index with odd number of arguments")

        index_dims = args[0::2]
        for dim in index_dims:
            if dim not in a.dims:
                raise ValueError("dim not in dims")

        index_slice = args[1::2]

        # check slices
        for slice_ in index_slice:

            # type must be slice or tuple
            if not isinstance(slice_, (slice, tuple, float, int)):
                raise ValueError("Invalid slice type")

            # if tuple, length must be two: (start, stop)
            if isinstance(slice_, tuple) and not len(slice_) in (1, 2):
                raise ValueError("tuple index must have one or two values")

        # convert tuple to slice
        updated_index_slice = []
        for dim, slice_ in zip(index_dims, index_slice):
            if isinstance(slice_, tuple):
                index = a.index(dim)
                if len(slice_) == 1:
                    start = np.argmin(np.abs(slice_[0] - a.get_coord(dim)))
                    updated_index_slice.append(slice(start, start + 1))
                else:
                    start = np.argmin(np.abs(slice_[0] - a.get_coord(dim)))
                    stop = np.argmin(np.abs(slice_[1] - a.get_coord(dim)))
                    if start == stop:
                        stop = start + 1
                    if stop < start:
                        start, stop = stop, start
                    updated_index_slice.append(slice(start, stop))
            elif isinstance(slice_, int):
                start = slice_
                if slice_ != -1:
                    updated_index_slice.append(slice(start, start + 1))
                else:
                    updated_index_slice.append(slice(slice_, None))
            elif isinstance(slice_, float):
                start = np.argmin(np.abs(slice_ - a.get_coord(dim)))
                updated_index_slice.append(slice(start, start + 1))
            else:
                updated_index_slice.append(slice_)

        index_slice_dict = dict(zip(index_dims, updated_index_slice))
        new_slices = [
            slice(None) if dim not in index_dims else index_slice_dict[dim]
            for dim in a.dims
        ]

        self.values[tuple(new_slices)] = new_values

    def copy(self):
        """Return deepcopy of dnpdata object

        Returns:
            deep copy of data object
        """
        return deepcopy(self)

    def cumulative_sum(self, dim):
        """Calculate Cumulative sum of dnpdata object

        Returns:
            cumulative sum of data object
        """
        a = self.copy()
        index = a.index(dim)
        a.values = a.values.cumsum(index)
        # NOTE: Add Error Propagation
        return a

    def merge_attrs(self, b):
        """Merge the given dictionaries

        Args:
            b (nddata_core): attributes to merge into object
        """

        for key in b.attrs:
            if key not in self.attrs:
                self.attrs[key] = b.attrs[key]
            else:
                if not self.attrs[key] is b.attrs[key]:
                    warnings.warn(
                        "attrs in two dictionarys contain different values, leaving original value:\n{}:{}".format(
                            key, self.attrs[key]
                        )
                    )

    def __len__(self):
        """Returns len(values) the length of the first dimension in values"""
        return len(self._values)

    @property
    def size(self):
        """Returns values.size. Total number of elements in numpy array."""
        return self._values.size

    def sort_dims(self):
        """Sort the dimensions"""
        sorted_order = sorted(range(len(self.dims)), key=lambda x: self.dims[x])

        self.coords.reorder_index(sorted_order)
        self._values = np.moveaxis(self._values, range(len(sorted_order)), sorted_order)

    def index(self, dim):
        """Find index of given dimension name

        Args:
            dim (str): Name of dimension to index

        Returns:
            int: Index value of dim
        """
        if dim in self.dims:
            return self.coords.index(dim)
        else:
            raise ValueError("%s not in %s" % (dim, self.dims))

    def __truediv__(self, b):
        if isinstance(b, ABCData):

            a, b = self.align(b)

            a.values = a.values / b.values

            # error propagation
            if a.error is not None and b.error is not None:
                error = abs(a.values) * np.sqrt(
                    (self.error / a.values) ** 2.0 + (b.error / a.values) ** 2.0
                )
            elif b.error is not None:
                a.error = b.error

            return a

        elif isinstance(b, _numerical_types):
            a = self.copy()
            a.values = a.values.__truediv__(b)
            return a
        else:
            raise TypeError("Cannot add type: {}".format(type(b)))

    def __rtruediv__(self, b):
        if isinstance(b, _numerical_types):
            a = self.copy()
            a.values = self.values.__rtruediv__(b)
            return a
        else:
            raise TypeError("Cannot add type: {}".format(type(b)))

    def __mul__(self, b):
        if isinstance(b, ABCData):
            a, b = self.align(b)

            a.values = a.values * b.values

            # error propagation
            if a.error is not None and b.error is not None:
                error = abs(a.values) * np.sqrt(
                    (self.error / a.values) ** 2.0 + (b.error / a.values) ** 2.0
                )
            elif b.error is not None:
                a.error = b.error

            return a

        elif isinstance(b, _numerical_types):
            a = self.copy()
            a.values = a.values.__mul__(b)
            return a
        else:
            raise TypeError("Cannot add type: {}".format(type(b)))

    __rmul__ = __mul__

    def __add__(self, b):
        if isinstance(b, ABCData):
            a, b = self.align(b)

            a.values = a.values + b.values

            # error propagation
            if a.error is not None and b.error is not None:
                a.error = np.sqrt(a.error**2.0 + b.error**2.0)
            elif b.error is not None:
                a.error = b.error

            return a

        elif isinstance(b, _numerical_types):
            a = self.copy()
            a.values = a.values.__add__(b)
            return a
        else:
            raise TypeError("Cannot add type: {}".format(type(b)))

    __radd__ = __add__

    def __sub__(self, b):
        if isinstance(b, ABCData):
            a, b = self.align(b)

            a.values = a.values - b.values

            # error propagation
            if a.error is not None and b.error is not None:
                a.error = np.sqrt(a.error**2.0 + b.error**2.0)
            elif b.error is not None:
                a.error = b.error

            return a

        elif isinstance(b, _numerical_types):
            a = self.copy()
            a.values = a.values.__sub__(b)
            return a
        else:
            raise TypeError("Cannot add type: {}".format(type(b)))

    def __rsub__(self, b):
        if isinstance(b, _numerical_types):
            a = self.copy()
            a.values = a.values.__rsub__(b)
            return a
        else:
            raise TypeError("Cannot add type: {}".format(type(b)))

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, b):
        if not isinstance(b, (int, complex, float, np.ndarray)):
            raise TypeError('Values must be type "numpy.ndarray" not %s' % type(b))
        if isinstance(b, (int, complex, float)):
            b = np.array(b)

        self._values = b

    @values.getter
    def values(self):
        return self._values

    @property
    def dims(self):
        return self.coords.dims

    @dims.setter
    def dims(self, b):
        self.coords.dims = b

    @property
    def coords(self):
        return self._coords

    @coords.setter
    def coords(self, b):
        self._coords = b

    @property
    def attrs(self):
        return self._attrs

    @attrs.setter
    def attrs(self, b):
        if isinstance(b, dict):
            self._attrs = b
        else:
            raise ValueError('attrs must be type "dict"')

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, b):
        if isinstance(b, (np.ndarray, type(None))):
            self._error = b
        else:
            raise ValueError('error must be type "numpy.ndarray"')

    def rename(self, dim, new_name):
        """Rename dim

        Args:
            dim (str): Name of dimension to rename
            new_name (str): New name for dim
        """
        if dim in self.dims:
            if isinstance(new_name, str):
                self.coords.rename(dim, new_name)
            else:
                raise TypeError(
                    'New dimension name must be type "str" not %s' % type(new_name)
                )
        else:
            raise ValueError("Dimension name %s is not in dims" % dim)

    def reorder(self, dims):
        """Reorder dimensions

        Args:
            dims (list): List of strings in new order

        """

        if not self._check_dims(dims):
            raise TypeError("New dims must be list of str with no duplicates")
        for dim in dims:
            if dim not in self.dims:
                raise ValueError("no such dimension: %s" % dim)

        # Add original dims to end, remove duplicates
        dims = list(OrderedDict.fromkeys(dims + self.dims))

        #        print('original dims', self.dims)
        #        print('new order of dims', dims)

        new_order = [dims.index(dim) for dim in self.dims]
        permutation_order = [self.dims.index(dim) for dim in dims]
        #        print('New Order', new_order)
        #        print('Permutation Order', permutation_order)

        self.coords.reorder(dims)
        #        print(self.coords)

        # Transpose values
        #        self.values = np.transpose(self.values, new_order)
        self.values = np.transpose(self.values, permutation_order)

    def __str__(self):
        return "values:\n{}\ndims:\n{}\ncoords:\n{}\nattrs:\n{}".format(
            self.values, self.dims, self.coords, self.attrs
        )

    def __repr__(self):
        return "nddata_core(values = {}, coords = {}, dims = {}, attrs = {})".format(
            repr(self.values), repr(self.dims), repr(self.coords), repr(self.attrs)
        )

    def squeeze(self, dim):
        """Remove length 1 axes"""
        a = self.copy()
        shape = a.shape

        remove_dims = [a.dims[x] for x in range(len(shape)) if shape[x] == 1]
        values = np.squeeze(a.values)

        if a.error is not None:
            a.error = np.squeeze(a.error)

        attrs = a.attrs

        for dim in remove_dims:
            out = a.coords.pop(dim)
            if dim not in attrs:
                attrs[dim] = np.array(out)
            else:
                warnings.warn("Attribute lost {}:{}".format(dim, out))

        return a

    def chunk(self, dim, new_dims, new_sizes):
        """
        .. note::
            This is a placeholder for a function that's not yet implemented

        Parameters
        ----------
        dim: str
            Assume that the dimension `dim` is a direct product of the
            dimensions given in `new_dims`, and chunk it out into those new
            dimensions.
        new_dims: list of str
            The new dimensions to generate.  Note that one of the elements of
            the list can be `dim` if you like.

            It's assumed that the ordering of `dim` is a direct product given
            in C-ordering (*i.e.* the inner dimensions are listed last and the
            outer dimensions are listed first -- here "inner" means that
            changes to the index of the inner-most dimension correspond to
            adjacent positions in memory and/or adjacent indeces in the
            original dimension that you are chunking)
        new_sizes: list of int
            sizes of the new dimensions`
        Returns
        -------
        self: nddata_core
            The new nddata object.
            Note that uniformly ascending or descending coordinates are manipulated in a rational way,
            *e.g.* `[1,2,3,4,5,6]` when chunked to a size of `[2,3]` will yield
            coordinates for the two new dimensions:
            `[1,4]` and `[0,1,2]`.
            Coordinates that are not uniformly ascending or descending will
            yield and error and must be manually modified by the user.
        """
        return NotImplemented

    def get_coord(self, dim):
        """Return coord corresponding to given dimension name

        Args:
            dim (str): Name of dim to retrieve coordinates from

        Returns:
            numpy.ndarray: array of coordinates

        """
        return self.coords[dim]

    @property
    def shape(self):
        """tuple: Shape of values"""
        return self.values.shape

    @property
    def dtype(self):
        """type: Values type"""
        return self.values.dtype

    def sum(self, dim):
        """Perform sum down given dimension

        Args:
            dim (str): Dimension to perform sum down

        """
        a = self.copy()
        index = a.index(dim)
        a.values = a.values.sum(index)
        if a.error is not None:
            a.error = a.error.std(index)
        a.coords.pop(dim)
        return a

    def align(self, b):
        """Align two data objects for numerical operations

        Args:
            b: Ojbect to align with self

        Returns:
            tuple: self and b aligned data objects
        """
        a = self.copy()
        b = b.copy()

        all_dims = list(OrderedDict.fromkeys(a.dims + b.dims))
        new_b_order = [dim for dim in all_dims if dim in b.dims]
        new_order = [b.index(dim) for dim in new_b_order]

        # create new dims where necessary
        values = a.values[
            tuple(slice(None) if dim in a.dims else None for dim in all_dims)
        ]

        # re-order
        old_order = list(range(len(new_order)))
        # re-order b values so they match order of all_dims
        values_b = np.moveaxis(b.values, new_order, old_order)
        # create new dims where necessary
        values_b = values_b[
            tuple(slice(None) if dim in b.dims else None for dim in all_dims)
        ]

        # Handle Error
        if a.error is not None:
            error = a.error[
                tuple(slice(None) if dim in a.dims else None for dim in all_dims)
            ]
        else:
            error = a.error
        if b.error is not None:
            error_b = np.moveaxis(b.values, new_order, old_order)
            error_b = error_b[
                tuple(slice(None) if dim in b.dims else None for dim in all_dims)
            ]
        else:
            error_b = b.error

        # check coords
        for dim in all_dims:
            if (dim in a.dims) and (dim in b.dims):
                if not np.allclose(a.get_coord(dim), b.get_coord(dim)):
                    raise ValueError("Coords do not match for dim: %s" % dim)

        # merge attrs
        a.merge_attrs(b)

        # coords
        coords = a.coords.copy()
        for dim in new_b_order:
            if dim not in a.dims:
                coords.append(dim, b.coords[dim])

        a.values = values
        a.coords = coords
        a.error = error

        b.values = values_b
        b.coords = coords
        b.error = error_b

        return a, b

    def __array__(self):
        return self.values

    def smoosh(self, old_dims, new_name):
        """
        .. note::
            Not yet implemented.

        `smoosh` does the opposite of `chunk` -- see :func`:~nddata_core.chunk`
        """
        return NotImplemented

    def sort(self, dim):
        """Sort the coords corresponding to the given dim in ascending order

        Args:
            dim (str): dimension to sort
        """

        sort_array = np.argsort(self.coords[dim])

        self.coords[dim] = self.coords[dim][sort_array]

        new_order = tuple(
            [slice(None) if dim != this_dim else sort_array for this_dim in self.dims]
        )

        self.values = self.values[new_order]

    def split(self, dim, new_dim, coord):
        """Split the dimension dim into"""

        if isinstance(coord, int):
            coord = np.arange(coord)

        # move dim to end of dims
        dims = self.dims
        dims.remove(dim)
        dims.append(dim)

        # reorder data with split dim at end
        self.reorder(dims)

        new_shape = list(self.coords.shape)
        new_shape[-1] = int(new_shape[-1] / coord.size)
        self.coords[dim] = self.coords[dim][0 : new_shape[-1]]

        new_shape += [coord.size]
        print(new_shape)

        self.values = self.values.reshape(new_shape)
        self.coords.append(new_dim, coord)

    def is_sorted(self, dim):
        """Determine if coords corresponding to give dim are sorted in ascending order
        Args:
            dim (str): Dimension to check if sorted

        Returns:
            bool: True if sorted, False otherwise.
        """
        return np.all(self.coords[dim][:-1] <= self.coords[dim][1:])

    @property
    def real(self):
        """DNPData: DNPData with real part of values"""
        a = self.copy()
        a.values = np.real(a.values)
        return a

    @property
    def imag(self):
        """DNPData: DNPData with imaginary part of values"""
        a = self.copy()
        a.values = np.imag(a.values)
        return a

    @property
    def abs(self):
        """DNPData: DNPData with absolute part of values"""
        a = self.copy()
        a.values = np.abs(a.values)
        return a

    def concatenate(self, b, dim):
        """Concatenate DNPData objects

        Args:
            b (DNPData): Data object to append to current data object
            dim (str): dimension to concatenate along
        """

        if not dim in b.dims:
            raise ValueError("dim does not exist")
        if not dim in self.dims:
            raise ValueError("dim does not exist")

        index = self.dims.index(dim)

        b.reorder(self.dims)

        self.values = np.concatenate((self.values, b.values), axis=index)

        self.coords[dim] = np.concatenate(
            (
                np.array(self.coords[dim]).reshape(-1),
                np.array(b.coords[dim]).reshape(-1),
            )
        )

    def new_dim(self, dim, coord):
        """Add new dimension with length 1

        Args:
            dim (str): Name of new dimension
            coord (int, float): New coord
        """
        self.coords.append(dim, np.r_[coord])
        self.values = np.expand_dims(self.values, -1)

    def maximum(self, dim):
        """Return max for given dim

        Args:
            dim (str): Dimension to take maximum along

        """
        a = self.copy()
        index = a.dims.index(dim)

        a.values = np.max(a.values, axis=index)
        a.coords.pop(dim)

        return a

    def argmax(self, dim):
        """Return value of coord at values maximum for given dim

        Args:
            dim (str): Dimension to perform operation along
        """
        a = self.copy()
        index = a.dims.index(dim)

        a.values = a.coords[dim][np.argmax(a.values, axis=index)]
        a.coords.pop(dim)

        return a

    def argmax_index(self, dim):
        """Return index of coord at values maximum for given dim

        Args:
            dim (str): Dimension to perform operation along
        """
        a = self.copy()
        index = a.dims.index(dim)

        a.values = np.argmax(a.values, axis=index)
        a.coords.pop(dim)

        return a

    def minimum(self, dim):
        """Return min for given dim

        Args:
            dim (str): Dimension to perform operation along
        """
        a = self.copy()
        index = a.dims.index(dim)

        a.values = np.min(a.values, axis=index)
        a.coords.pop(dim)

        return a

    def argmin(self, dim):
        """Return value of coord at values minimum for given dim

        Args:
            dim (str): Dimension to perform operation along
        """
        a = self.copy()
        index = a.dims.index(dim)

        a.values = a.coords[dim][np.argmin(a.values, axis=index)]
        a.coords.pop(dim)

        return a

    def argmin_index(self, dim):
        """Return index of coord at values minimum for given dim

        Args:
            dim (str): Dimension to perform operation along
        """
        a = self.copy()
        index = a.dims.index(dim)

        a.values = np.argmin(a.values, axis=index)
        a.coords.pop(dim)

        return a

    @property
    def ndim(self):
        """str: Number of dimensions"""
        return self.values.ndim

    def unfold(self, dim):
        """Unfold ND data to 2d data

        Args:
            dim (str): Dimension to make first (length N), all other dimensions unfolded so that values has shape (N x M)

        """

        folded_order = self.dims  # Original order of dims
        self.reorder([dim])  # Move dim to first dimension
        folded_shape = self.values.shape
        align_dim_length = folded_shape[0]  # length of dimension to align down
        self.values = self.values.reshape(align_dim_length, -1)  # Reshape to 2d
        self.attrs["folded_shape"] = folded_shape
        self.attrs["folded_order"] = folded_order

    def fold(self):
        """Fold 2d data to original ND shape"""

        if "folded_shape" in self.attrs:
            original_shape = self.attrs["folded_shape"]
            self.values = self.values.reshape(original_shape)
            self.attrs.pop("folded_shape")
        if "folded_order" in self.attrs:
            self.reorder(self.attrs["folded_order"])
            self.attrs.pop("folded_order")



"""Functions to import Bruker EPR data"""

import numpy as np
import os
from .. import DNPData
import warnings


def import_bes3t(path):
    """
    Import Bruker BES3T data and return dnpdata object

    Args:
        path (str) : Path to either .DSC or .DTA file

    Returns:
        bes3t_data (object) : DNPData object containing Bruker BES3T data
    """

    pathexten = os.path.splitext(path)[1]
    path = os.path.splitext(path)[0]
    path_xgf = "none"
    path_ygf = "none"
    path_zgf = "none"
    if pathexten == ".DSC" or pathexten == ".DTA":
        path_dsc = path + ".DSC"
        path_dta = path + ".DTA"
        if os.path.isfile(path + ".XGF"):
            path_xgf = path + ".XGF"
        if os.path.isfile(path + ".YGF"):
            path_ygf = path + ".YGF"
        if os.path.isfile(path + ".ZGF"):
            path_zgf = path + ".ZGF"

    elif pathexten == ".XGF":
        path_xgf = path + ".XGF"
        path_dsc = path + ".DSC"
        path_dta = path + ".DTA"
        if os.path.isfile(path + ".YGF"):
            path_ygf = path + ".YGF"
        if os.path.isfile(path + ".ZGF"):
            path_zgf = path + ".ZGF"

    elif pathexten == ".YGF":
        path_ygf = path + ".YGF"
        path_dsc = path + ".DSC"
        path_dta = path + ".DTA"
        if os.path.isfile(path + ".XGF"):
            path_xgf = path + ".XGF"
        if os.path.isfile(path + ".ZGF"):
            path_zgf = path + ".ZGF"

    elif pathexten == ".ZGF":
        path_zgf = path + ".ZGF"
        path_dsc = path + ".DSC"
        path_dta = path + ".DTA"
        if os.path.isfile(path + ".XGF"):
            path_xgf = path + ".XGF"
        if os.path.isfile(path + ".YGF"):
            path_ygf = path + ".YGF"

    else:
        raise TypeError("data file must be .DTA, .DSC, .YGF, or .ZGF")

    params = load_dsc(path_dsc)
    values, dims, coords, attrs = load_dta(
        path_dta, path_xgf, path_ygf, path_zgf, params
    )

    bes3t_data = DNPData(values, dims, coords, attrs)

    return bes3t_data


def load_dsc(path):
    """
    Import contents of .DSC file

    Args:
        path (str) : Path to .DSC file

    Returns:
        params (dict) : dictionary of parameters
    """

    file_opened = open(path, "r")
    dscfile_contents = file_opened.readlines()
    file_opened.close()

    params = {}
    sweep_domain = None
    for ix in range(len(dscfile_contents)):
        try:
            par = dscfile_contents[ix].rstrip("\t").rstrip("\n")
            if "MWFQ" in par:
                params["frequency"] = float(par.replace("MWFQ", "").strip()) / 1e9
            elif "Power" in par and "Atten" not in par:
                params["power"] = float(
                    par.replace("Power", "").replace("mW", "").strip()
                )
            elif "PowerAtten" in par:
                params["attenuation"] = int(
                    float(par.replace("PowerAtten", "").replace("dB", "").strip())
                )
            elif "Attenuation" in par:
                params["pulse_attenuation"] = int(
                    float(par.replace("Attenuation", "").replace("dB", "").strip())
                )
            elif "CenterField" in par:
                params["center_field"] = float(
                    par.replace("CenterField", "").replace("G", "").strip()
                )
            elif "ConvTime" in par:
                params["conversion_time"] = float(
                    par.replace("ConvTime", "").replace("ms", "").strip()
                )
            elif "TimeConst" in par:
                params["time_constant"] = float(
                    par.replace("TimeConst", "").replace("ms", "").strip()
                )
            elif "ModAmp" in par:
                params["modulation_amplitude"] = float(
                    par.replace("ModAmp", "").replace("G", "").strip()
                )
            elif "ModFreq" in par:
                params["modulation_frequency"] = float(
                    par.replace("ModFreq", "").replace("kHz", "").strip()
                )
            elif "NbScansDone" in par:
                params["nscans"] = int(par.replace("NbScansDone", "").strip())
            elif "Temperature" in par:
                params["temperature"] = float(
                    par.replace("Temperature", "").replace("K", "").strip()
                )
            elif "XNAM" in par:
                sweep_domain = par.replace("XNAM", "").replace("'", "").strip()
            elif "XUNI" in par:
                params["x_unit"] = par.replace("XUNI", "").replace("'", "").strip()
            elif "XPTS" in par:
                params["x_points"] = int(par.replace("XPTS", "").strip())
            elif "XMIN" in par:
                params["x_min"] = float(par.replace("XMIN", "").strip())
            elif "XWID" in par:
                params["x_width"] = float(par.replace("XWID", "").strip())
            elif "XTYP" in par:
                xtyp = par.replace("XTYP", "").strip()
                if xtyp == "IGD":
                    params["x_type"] = "nonlinear"
                elif xtyp == "IDX":
                    params["x_type"] = "linear"

            elif "XFMT" in par:
                params["x_format"] = _return_data_type(par, "XFMT")

            elif "YNAM" in par:
                y_domain = par.replace("YNAM", "").replace("'", "").strip()
            elif "YUNI" in par:
                params["y_unit"] = par.replace("YUNI", "").replace("'", "").strip()
            elif "YPTS" in par:
                params["y_points"] = int(par.replace("YPTS", "").strip())
            elif "YMIN" in par:
                params["y_min"] = float(par.replace("YMIN", "").strip())
            elif "YWID" in par:
                params["y_width"] = float(par.replace("YWID", "").strip())
            elif "YTYP" in par:
                ytyp = par.replace("YTYP", "").strip()
                if ytyp == "IGD":
                    params["y_type"] = "nonlinear"
                elif ytyp == "IDX":
                    params["y_type"] = "linear"

            elif "YFMT" in par:
                params["y_format"] = _return_data_type(par, "YFMT")

            elif "ZNAM" in par:
                z_domain = par.replace("ZNAM", "").replace("'", "").strip()
            elif "ZUNI" in par:
                params["z_unit"] = par.replace("ZUNI", "").replace("'", "").strip()
            elif "ZPTS" in par:
                params["z_points"] = int(par.replace("ZPTS", "").strip())
            elif "ZMIN" in par:
                params["z_min"] = float(par.replace("ZMIN", "").strip())
            elif "ZWID" in par:
                params["z_width"] = float(par.replace("ZWID", "").strip())
            elif "ZTYP" in par:
                ztyp = par.replace("ZTYP", "").strip()
                if ztyp == "IGD":
                    params["z_type"] = "nonlinear"
                elif ztyp == "IDX":
                    params["z_type"] = "linear"

            elif "ZFMT" in par:
                params["z_format"] = _return_data_type(par, "ZFMT")

            elif "IRFMT" in par:
                params["real_format"] = _return_data_type(par, "IRFMT")

            elif "IIFMT" in par:
                params["imag_format"] = _return_data_type(par, "IIFMT")

            elif "IKKF" in par:
                params["data_type"] = par.replace("IKKF", "").strip()
            elif "BSEQ" in par:
                params["endian"] = par.replace("BSEQ", "").strip()

        except ValueError:
            continue

    if sweep_domain == "Time" and int(params["attenuation"]) == 60:
        params.pop("attenuation", None)
        params.pop("power", None)
    elif sweep_domain == "Time" and int(params["pulse_attenuation"]) == 60:
        params["pulse_attenuation"] = params["attenuation"]
        params.pop("attenuation", None)

    if all(
        [
            y not in params.keys()
            for y in ["real_format", "imag_format", "x_format", "y_format", "z_format"]
        ]
    ):
        raise TypeError(
            "data format unknown. None of IRFMT, IIFMT, XFMT, YFMT, ZFMT found in DSC"
        )

    for x in ["x", "y", "z"]:
        if x + "_format" not in params.keys():
            if params["data_type"] == "REAL":
                params[x + "_format"] = params["real_format"]
            elif params["data_type"] == "CPLX":
                params[x + "_format"] = params["imag_format"]

    return params


def load_dta(path_dta, path_xgf=None, path_ygf=None, path_zgf=None, params={}):
    """
    Import data from .DTA file. Uses .DSC and .XGF, .YGF, or .ZGF files if they exists

    Args:
        path_dta (str) : Path to .DTA file
        path_xgf (str) : path to .XGF file for 1D data with nonlinear axis, "none" otherwise
        path_ygf (str) : path to .YGF file for 2D data, "none" if 1D or linear y axis
        path_zgf (str) : path to .ZGF file for 3D data, "none" if 1D/2D or linear z axis
        params (dict) : dictionary of parameters

    Returns:
        abscissa (ndarray) : coordinates for spectrum or spectra
        spec (ndarray) : spectrum for 1D or spectra for 2D
        params (dict) : updated dictionary of parameters
        dims (list) : dimensions
    """

    dta_dtype = np.dtype(params["x_format"]).newbyteorder(params["endian"])
    file_opened = open(path_dta, "rb")
    file_bytes = file_opened.read()
    file_opened.close()
    spec = np.frombuffer(file_bytes, dtype=dta_dtype)
    if params["x_type"] == "nonlinear":
        abscissa = [
            load_gf_files(
                path_xgf,
                axis_type=params["x_type"],
                axis_format=params["x_format"],
                axis_points=params["x_points"],
                axis_min=params["x_min"],
                axis_width=params["x_width"],
                endian=params["endian"],
            )
        ]
    elif params["x_type"] == "linear":
        abscissa = [
            np.linspace(
                params["x_min"],
                params["x_min"] + params["x_width"],
                params["x_points"],
            )
        ]

    if "x_unit" in params.keys() and params["x_unit"] in ["G", "T"]:
        if params["x_unit"] == "G":
            abscissa = [x / 10 for x in abscissa]
        elif params["x_unit"] == "T":
            abscissa = [x * 1000 for x in abscissa]
        dims = ["B0"]
    else:
        dims = ["t2"]

    if params["data_type"] == "CPLX":
        spec = spec.astype(dtype=params["imag_format"]).view(dtype=np.dtype("complex"))
    elif params["data_type"] == "REAL":
        spec = spec.astype(dtype=params["real_format"]).view()

    if (
        "z_points" not in params.keys()
        or ("z_points" in params.keys() and params["z_points"] == 1)
    ) and ("y_points" in params.keys() and params["y_points"] != 1):
        spec = np.reshape(spec, (params["x_points"], params["y_points"]), order="F")

        dims.append("t1")

        abscissa.append(
            load_gf_files(
                path_ygf,
                axis_type=params["y_type"],
                axis_format=params["y_format"],
                axis_points=params["y_points"],
                axis_min=params["y_min"],
                axis_width=params["y_width"],
                endian=params["endian"],
            )
        )

    elif "z_points" in params.keys() and params["z_points"] != 1:
        spec = np.reshape(
            spec,
            (params["x_points"], params["y_points"]),
            params["z_points"],
            order="F",
        )

        abscissa.append(
            load_gf_files(
                path_ygf,
                axis_type=params["y_type"],
                axis_format=params["y_format"],
                axis_points=params["y_points"],
                axis_min=params["y_min"],
                axis_width=params["y_width"],
                endian=params["endian"],
            )
        )
        dims.append("t1")

        abscissa.append(
            load_gf_files(
                path_zgf,
                axis_type=params["z_type"],
                axis_format=params["z_format"],
                axis_points=params["z_points"],
                axis_min=params["z_min"],
                axis_width=params["z_width"],
                endian=params["endian"],
            )
        )
        dims.append("t0")

    params = {
        x: params[x]
        for x in params.keys()
        if x
        not in [
            "endian",
            "x_format",
            "x_type",
            "x_points",
            "x_min",
            "x_width",
            "y_format",
            "y_type",
            "y_points",
            "y_min",
            "y_width",
            "z_format",
            "z_type",
            "z_points",
            "z_min",
            "z_width",
            "real_format",
            "imag_format",
            "data_type",
        ]
    }

    return spec, dims, abscissa, params


def load_gf_files(
    path,
    axis_type="",
    axis_format="",
    axis_points=1,
    axis_min=1,
    axis_width=1,
    endian="",
):
    """
    Import data from .XGF, .YGF, or .ZGF files

    Args:
        path (str) : Path to ._GF file
        axis_type (str) : linear or nonlinear
        axis_format (str) : format of file data
        axis_points (int) : number of points in axis
        axis_min (float) : minimum value of axis
        axis_width (float) : total width of axis
        endian (float) : endian of data

    Returns:
        abscissa (ndarray) : axis coordinates
    """

    if path != "none":
        if axis_type == "linear":
            warnings.warn("axis format is linear, confirm that the axis is correct")
        gf_type = np.dtype(axis_format).newbyteorder(endian)
        file_opened = open(path, "rb")
        file_bytes = file_opened.read()
        file_opened.close()
        abscissa = np.frombuffer(file_bytes, dtype=axis_format)
    elif path == "none":
        if axis_type == "nonlinear":
            warnings.warn(
                "axis is nonlinear, confirm that file is not needed and the axis is correct"
            )
        abscissa = np.linspace(axis_min, axis_min + axis_width, axis_points)
    else:
        warnings.warn("axis format not supported, axis is only indexed")
        abscissa = np.array([range(axis_points)])

    return abscissa


def _return_data_type(par, key):
    fmt = par.replace(key, "").strip()
    if fmt == "D":
        return "float64"
    elif fmt == "F":
        return "float32"
    elif fmt == "C":
        return "int8"
    elif fmt == "S":
        return "int16"
    elif fmt == "I":
        return "int32"



import unittest
import numpy as np
from numpy.testing import assert_array_equal as assertArrayEqual
from .testing import get_gauss_3d
import dnplab as dnp


class dnpDataTester(unittest.TestCase):
    def setUp(self):
        self.x, self.y, self.z, self.gauss_3d = get_gauss_3d(0.1)
        self.data = dnp.DNPData(
            self.gauss_3d, ["x", "y", "z"], [self.x, self.y, self.z]
        )

    def test_DNPData(self):
        assertArrayEqual(self.data.coords["x"], self.x)
        assertArrayEqual(self.data.dims, ["x", "y", "z"])

    def test_coords_get_set(self):
        self.data.new_dim("r", np.r_[0:10])
        assertArrayEqual(self.data.dims, ["x", "y", "z", "r"])
        self.data.rename("r", "s")
        assertArrayEqual(self.data.dims, ["x", "y", "z", "s"])

    def test_coords_sort_reorder(self):
        self.data.reorder(["y", "z", "x"])
        assertArrayEqual(self.data.dims, ["y", "z", "x"])
        assertArrayEqual(self.data.coords["z"], self.z)
        self.data.sort_dims()
        assertArrayEqual(self.data.dims, ["x", "y", "z"])
        assertArrayEqual(self.data.coords["z"], self.z)


if __name__ == "__main__":
    pass



import numpy as np
from struct import unpack
from .. import DNPData


def import_delta(path):
    """Import Delta data and return DNPData object

    Args:
        path (str) : Path to .jdf file

    Returns:
        delta_data (object) : DNPData object containing Delta data
    """

    pars = import_delta_pars(path)
    values, dims, coords, attrs = import_delta_data(path, pars)

    delta_data = DNPData(values, dims, coords, attrs)

    return delta_data


def import_delta_pars(path):
    """Import parameter fields of Delta data

    Args:
        path (str) : Path to .jdf file

    Returns:
        params (dict) : dictionary of parameter fields and values
    """

    file_opened = open(path, "rb")
    file_contents = file_opened.readlines()
    file_opened.close()
    params = {}
    for ix in range(len(file_contents)):
        try:
            file_contents[ix] = str(file_contents[ix], "utf8")
            if "=" in file_contents[ix]:
                new_line = file_contents[ix].split("=")
                new_name = new_line[0].replace("/", "").strip()
                if "@" in new_name or new_name == "":
                    pass
                else:
                    params[new_name] = (
                        new_line[1]
                        .replace(">", "")
                        .replace(";", "")
                        .replace("?", "")
                        .strip()
                    )
        except UnicodeDecodeError:
            pass

    return params


def import_delta_data(path, params):
    """Import spectrum or spectra of Delta data

    Args:
        path (str) : Path to .jdf file
        params (dict) : dictionary of parameters

    Returns:
        y_data (ndarray) : spectrum or spectra if >1D
        abscissa (list) : coordinates of axes
        dims (list) : axes names
        params (dict) : updated dictionary of parameters
    """

    if not params:
        params = {}

    file_opened = open(path, "rb")
    file_contents = file_opened.read(1296)
    file_opened.close()

    num_dims = [
        unpack(">B", file_contents[12 + ix : 13 + ix])[0] for ix in range(0, 1, 1)
    ][0]

    params["nmr_frequency"] = [
        unpack(">d", file_contents[1064 + ix : 1072 + ix])[0] for ix in range(0, 64, 8)
    ][
        0
    ] * 1e6  # convert from MHz to Hz, is this always in Hz?
    axes_units = [
        unpack(">B", file_contents[32 + ix : 33 + ix])[0] for ix in range(1, 16, 2)
    ][:num_dims]
    params["units"] = []
    for ix in range(num_dims):
        if axes_units[ix] == 1:
            params["units"].append("abundance")
        elif axes_units[ix] == 13:
            params["units"].append("Hz")
        elif axes_units[ix] == 26:
            params["units"].append("ppm")
        elif axes_units[ix] == 27:
            params["units"].append("rad")
        elif axes_units[ix] == 28:
            params["units"].append("s")
        else:
            params["units"].append("indexed")

    endian = [unpack(">B", file_contents[8 + ix : 9 + ix])[0] for ix in range(0, 1, 1)][
        0
    ]
    if endian == 0:
        endian = ">d"
    elif endian == 1:
        endian = "<d"
    else:
        raise UnicodeTranslateError("Failed to determine endianness")

    num_pts = [
        unpack(">I", file_contents[176 + ix : 180 + ix])[0] for ix in range(0, 32, 4)
    ]
    axis_type = [
        unpack(">B", file_contents[24 + ix : 25 + ix])[0] for ix in range(0, 8, 1)
    ][:num_dims]
    axes_start = [
        unpack(">d", file_contents[272 + ix : 280 + ix])[0] for ix in range(0, 64, 8)
    ][:num_dims]
    axes_stop = [
        unpack(">d", file_contents[336 + ix : 344 + ix])[0] for ix in range(0, 64, 8)
    ][:num_dims]
    abscissa = []
    for ix in range(num_dims):
        abscissa.append(np.linspace(axes_start[ix], axes_stop[ix], num_pts[ix]))

    data_start = [
        unpack(">I", file_contents[1284 + ix : 1288 + ix])[0] for ix in range(0, 4, 4)
    ][0]

    file_opened = open(path, "rb")
    file_opened.seek(data_start)
    if num_dims == 2 and axis_type[0] == 3 and axis_type[1] == 3:
        read_pts = np.prod(num_pts) * 4
    else:
        read_pts = np.prod(num_pts) * 2
    data = np.fromfile(file_opened, endian, read_pts)
    file_opened.close()

    if num_dims == 1:
        if axis_type[0] == 1:
            y_data = data
        elif axis_type[0] == 3 or axis_type[0] == 4:
            y_data = np.split(data, 2)[0] - 1j * np.split(data, 2)[1]
        else:
            raise TypeError("Data format not recognized")

        dims = ["t2"]

    elif num_dims == 2:
        if axis_type[0] == 4 or (axis_type[0] == 3 and axis_type[1] == 1):
            data_folded = np.split(data, 2)[0] - 1j * np.split(data, 2)[1]
            data_shaped = np.reshape(
                data_folded, [int(num_pts[0] / 4), int(num_pts[1] / 4), 4, 4], order="F"
            )
            y_data = np.concatenate(np.concatenate(data_shaped, 1), 1)
        elif axis_type[0] == 3 and axis_type[1] == 3:
            data_folded = [
                np.split(data, 4)[0] - 1j * np.split(data, 4)[1],
                np.split(data, 4)[2] - 1j * np.split(data, 4)[3],
            ]
            for idx in enumerate(data_folded):
                data_shaped[idx] = np.reshape(
                    data_folded[idx],
                    [int(num_pts[0] / 32), int(num_pts[1] / 32), 32, 32],
                    order="F",
                )
                y_data[idx] = np.concatenate(np.concatenate(data_shaped[idx], 1), 1)
        else:
            raise ValueError("Data format not recognized")

        dims = ["t2", "t1"]

    else:
        raise TypeError("Only 1D or 2D are supported")

    return y_data, dims, abscissa, params
